#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import itertools
from email import utils as email_utils

from attr import attrib
from attr import attrs
from attr import Factory
from attr import fields_dict

from debian_inspector import deb822
from debian_inspector import debcon

"""
Utilities to parse Debian machine readable copyright files (aka. dep5)
https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
"""


@attrs
class LicenseField(debcon.FieldMixin):
    name = attrib(default=None)
    text = attrib(default=None)

    @classmethod
    def from_value(cls, value):
        if isinstance(value, cls):
            return value
        lic = debcon.DescriptionField.from_value(value)
        syn = lic.synopsis
        if syn:
            syn = syn.lstrip()
        syn = lic.synopsis

        text = lic.text
        if text:
            text = text.lstrip()
        return cls(name=syn, text=text)

    def dumps(self, **kwargs):
        lic = debcon.DescriptionField(self.name, self.text)
        return lic.dumps(**kwargs).strip()

    def has_doc_reference(self):
        """
        Return True if this license contains a reference to a Debian shared
        license file in the /usr/share/common-licenses directory.
        """
        return self.text and '/usr/share/common-licenses' in self.text


@attrs
class CopyrightStatementField(debcon.FieldMixin):
    """
    Conventionally (but not in the spec) each line in a copyright is a space-
    separated tuple of (year range, holder). If it cannot be parsed, the holder
    contains all text.
    This field represents one line, e.g. one statememt.
    """
    # TODO: add line tracking
    holder = attrib()
    year_range = attrib(default=None)

    @classmethod
    def from_value(cls, value):
        if isinstance(value, cls):
            return value
        value = value or ''
        if isinstance(value, bytes):
            value = value.decode('utf-8')
        value = ' '.join(value.split())
        year_range, _, holder = value.partition(' ')
        year_range = year_range.strip()
        holder = holder.strip()
        if not is_year_range(year_range):
            holder = value
            year_range = None
        return cls(holder=holder, year_range=year_range)

    def dumps(self, **kwargs):
        cop = self.holder
        if self.year_range:
            cop = '{} {}'.format(self.year_range, cop)
        return cop.strip()


def is_year_range(text):
    """
    Return True if `text` is a year range.
    """
    if not text:
        return
    if all(c.isdigit() for c in text):
        return True

    digit_punct = set('''!"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~ 1234567890''')
    if all(c in digit_punct for c in text) and any(c.isdigit() for c in text):
        return True


@attrs
class CopyrightField(debcon.FieldMixin):
    """
    This represents a single "Copyright:" field which is a plain formatted text
    but is conventionally a list of copyrights statements one per line
    """
    statements = attrib(default=Factory(list))

    @classmethod
    def from_value(cls, value):
        if isinstance(value, cls):
            return value
        statements = []
        if value:
            statements = [
                CopyrightStatementField.from_value(v)
                for v in debcon.line_separated(value)]
        return cls(statements=statements)

    def dumps(self, **kwargs):
        dumped = [
            s.dumps(**kwargs) if hasattr(s, 'dumps') else str(s)
            for s in self.statements
        ]
        return '\n           '.join(dumped).strip()


@attrs
class MaintainerField(debcon.FieldMixin):
    """
    https://www.debian.org/doc/debian-policy/ch-controlfields#s-f-maintainer
    5.6.2. Maintainer
    """
    name = attrib()
    email_address = attrib(default=None)

    @classmethod
    def from_value(cls, value):
        if isinstance(value, cls):
            return value
        name = email_address = None
        if value:
            value = value.strip()
            name, email_address = email_utils.parseaddr(value)
            if not name:
                name = value
                email_address = None
            return cls(name=name, email_address=email_address)

    def dumps(self, **kwargs):
        name = self.name
        if self.email_address:
            name = '{} <{}>'.format(name, self.email_address)
        return name.strip()


@attrs
class BaseParagraph(debcon.FieldMixin):
    """
    Base paragraph with line numbers tracking.
    """

    # a mapping of {field_name: (start_line, end_line)} for each field
    line_numbers_by_field = attrib(default=Factory(dict))

    def get_field_names(self):
        """
        Return a list of field names defined on this paragraph.
        """
        return fields_dict(self.__class__)

    def get_field_line_numbers(self, field_name):
        """
        Return a tuple of (start_line, end_line) for the ``field_name`` field.
        """
        return self.line_numbers_by_field[field_name]

    def get_first_last_line_numbers(self):
        """
        Return a tuple of (first line, last line) number of this paragraph.
        """
        values = list(self.line_numbers_by_field.values())
        if values:
            starts, ends = list(zip(*values))

            if len(values) > 1:
                return min(starts), max(ends)
            elif len(values) == 1:
                return starts[0], ends[0]

        return 1, 1

    @classmethod
    def from_fields(cls, fields, all_extra=False):
        """
        Return a paragraph built from a list of Deb822Field.
        If ``all_extra`` is True, treat all data as "extra_data".
        """

        if all_extra:
            # if everything is "extra_data" this means there are no known names.
            known_names = set()
        else:
            known_names = set(fields_dict(cls))

        para_data = {}
        para_data['extra_data'] = extra_data = {}
        para_data['line_numbers_by_field'] = line_numbers_by_field = {}

        duplicated_field_name_suffix = 1
        seen_names = set()
        for field in fields:
            value = field.text

            if not value and not value.strip():
                continue

            name = field.name.replace('-', '_')

            # If there are duplicated fields, we keep them all, but rename them
            # with a number suffix; they will go in the extra_data mapping.
            if name in seen_names:
                name = f'{name}_{duplicated_field_name_suffix}'
                duplicated_field_name_suffix += 1
            seen_names.add(name)

            if name in known_names:
                mapping = para_data
            else:
                mapping = extra_data
            assert name not in mapping

            # we only strip leading spaces, including a possible first empty line
            mapping[name] = value.lstrip()

            start_line = field.start_line
            if value.startswith('\n'):
                start_line += 1
            line_numbers_by_field[name] = (start_line, field.end_line,)

        try:
            return cls(**para_data)
        except Exception as e:
            raise Exception(cls, para_data) from e

    @classmethod
    def from_dict(cls, data):
        assert isinstance(data, dict)
        known_names = set(fields_dict(cls))
        known_data = {}
        known_data['extra_data'] = extra_data = {}
        for key, value in data.items():
            key = key.replace('-', '_')
            if value:
                if isinstance(value, list):
                    value = '\n'.join(value)
                if key in known_names:
                    known_data[key] = value
                else:
                    extra_data[key] = value

        return cls(**known_data)


    def to_dict(self, with_extra_data=True, with_lines=False):
        data = {}

        for name in fields_dict(self.__class__):
            if name in ('extra_data' , 'line_numbers_by_field'):
                continue

            value = getattr(self, name)
            if value:
                if hasattr(value, 'dumps'):
                    value = value.dumps()
                data[name] = value

        if with_extra_data:
            for name, value in getattr(self, 'extra_data', {}).items():
                if value:
                    # always treat these extra values as formatted
                    value = value and debcon.as_formatted_text(value)
                data[name] = value

        if with_lines:
            data['line_numbers_by_field'] = self.line_numbers_by_field

        return data

    def dumps(self, **kwargs):
        text = []
        for name, value in self.to_dict().items():
            if value and value.strip():
                name = name.replace('_', '-')
                name = debcon.normalize_control_field_name(name)
                if value.startswith(' '):
                    value = value[1:]
                text.append('{}: {}'.format(name, value))
        return '\n'.join(text).strip()

    def is_empty(self):
        """
        Return True if all fields are empty
        """
        return not any(self.to_dict().values())

    def has_extra_data(self):
        return bool(getattr(self, 'extra_data', False))


@attrs
class CatchAllParagraph(BaseParagraph):
    """
    A catch-all paragraph: everything is fed to the extra_data. Every field is
    treated as formatted text.
    """
    extra_data = attrib(default=Factory(dict))

    @classmethod
    def from_fields(cls, fields):
        return super(CatchAllParagraph, cls).from_fields(
            fields=fields,
            all_extra=True,
        )

    def is_all_unknown(self):
        """
        Return True if this is an "unknown" field.
        We use the "unknown" field name for things that do not have a name.
        """
        return all(k.startswith('unknown') for k in self.to_dict())

    def is_valid(self, strict=False):
        if strict:
            return False
        return not self.is_all_unknown()


@attrs
class CopyrightHeaderParagraph(BaseParagraph):
    """
    The header paragraph.

    https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/#header-paragraph
    """
    # Default should be:
    # https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
    # but we do not know yet if this a structured machine-readable format
    format = debcon.SingleLineField.attrib(default=None)

    upstream_name = debcon.SingleLineField.attrib(default=None)
    # TODO: each may be a Maintainer
    upstream_contact = debcon.LineSeparatedField.attrib(default=None)

    source = debcon.FormattedTextField.attrib(default=None)
    disclaimer = debcon.FormattedTextField.attrib(default=None)
    copyright = CopyrightField.attrib(default=None)
    license = LicenseField.attrib(default=None)
    comment = debcon.FormattedTextField.attrib(default=None)

    # This field is not yet official but seen in use. See:
    # https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=685506
    files_excluded = debcon.AnyWhiteSpaceSeparatedField.attrib(default=None)

    # This is an overflow of extra unknown fields for this paragraph
    extra_data = attrib(default=Factory(dict))

    def is_valid(self, strict=False):
        valid = is_machine_readable_copyright(self.format.value)
        if strict:
            valid = valid and not self.has_extra_data()
        return valid


def is_machine_readable_copyright(text):
    """
    Return True if a text is for a machine-readable copyright format.
    """
    return text and text[:100].lower().startswith((
        'format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0',
        'format: http://www.debian.org/doc/packaging-manuals/copyright-format/1.0',
    ))


@attrs
class CopyrightFilesParagraph(BaseParagraph):
    """
    A "files" paragraph with files, copyright, license and comment fields.

    https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/#files-paragraph
    """
    files = debcon.AnyWhiteSpaceSeparatedField.attrib(default=None)
    copyright = CopyrightField.attrib(default=None)
    license = LicenseField.attrib(default=None)
    comment = debcon.FormattedTextField.attrib(default=None)

    # this is an overflow of extra unknown fields for this paragraph
    extra_data = attrib(default=Factory(dict))

    def dumps(self, **kwargs):
        if self.is_empty():
            return 'Files: '
        else:
            return BaseParagraph.dumps(self)

    def is_empty(self):
        """
        Return True if this is empty.
        """
        return not any([
            self.files.values,
            self.license.name,
            self.license.text,
            self.comment.text,
            self.copyright.statements,
            self.extra_data,
        ])

    def is_valid(self, strict=False):
        valid = (
            self.files.values
            and self.copyright.statements
            and self.license.name or self.license.text)
        if strict:
            valid = valid and not self.has_extra_data()
        return valid


@attrs
class CopyrightLicenseParagraph(BaseParagraph):
    """
    A standalone license paragraph with license and comment fields, but no files.

    https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/#stand-alone-license-paragraph
    """
    license = LicenseField.attrib(default=None)
    comment = debcon.FormattedTextField.attrib(default=None)

    # this is an overflow of extra unknown fields for this paragraph
    extra_data = attrib(default=Factory(dict))

    def is_empty(self):
        """
        Return True if this is empty (e.g. was crated only because of a
        'License:' empty field.
        """
        return not any([
            self.extra_data,
            self.comment.text,
            self.license.name,
            self.license.text,
        ])

    def dumps(self, **kwargs):
        if self.is_empty():
            return 'License: '
        else:
            return BaseParagraph.dumps(self)

    def is_valid(self, strict=False):
        valid = self.license.name or (self.license.name and self.license.text)
        if strict:
            valid = valid and not self.has_extra_data()
        return valid


@attrs
class DebianCopyright(object):
    """
    A machine-readable debian copyright file.
    See https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
    """
    paragraphs = attrib(default=Factory(list))

    def __attrs_post_init__(self, *args, **kwargs):
        self.merge_contiguous_unknown_paragraphs()
        self.fold_contiguous_empty_license_followed_by_unknown()

    @classmethod
    def from_text(cls, text):
        fields_groups = deb822.get_paragraphs_as_field_groups(text)
        return cls.from_fields_groups(fields_groups)

    @classmethod
    def from_file(cls, location):
        fields_groups = deb822.get_paragraphs_as_field_groups_from_file(location)
        return cls.from_fields_groups(fields_groups)

    @classmethod
    def from_fields_groups(cls, fields_groups):
        """
        Return a DebianCopyright from a ``fields_groups`` list of list of
        Deb822Field.
        """
        collected_paragraphs = []
        for fields in fields_groups:
            field_names = set([hf.name for hf in fields])

            if 'format'in field_names or 'format-specification' in field_names:
                # let's be flexible and assume that we have a copyright file
                # header if some format field is there
                cp = CopyrightHeaderParagraph.from_fields(fields)

            elif 'files' in field_names:
                # do we have a "files"? this is a file fields
                cp = CopyrightFilesParagraph.from_fields(fields)

            elif 'license' in field_names:
                cp = CopyrightLicenseParagraph.from_fields(fields)

            else:
                # we catch all the rest as junk to be flexible and miss nothing
                cp = CatchAllParagraph.from_fields(fields)

            collected_paragraphs.append(cp)

        return cls(collected_paragraphs)

    def dumps(self, **kwargs):
        dumped = [p.dumps(**kwargs) for p in self.paragraphs]
        dumped = '\n\n'.join(dumped)
        return dumped + '\n'

    def to_dict(self, with_lines=False):
        return {
            'paragraphs': [p.to_dict(with_lines=with_lines) for p in self.paragraphs]
        }

    def get_header(self):
        """
        Return the header paragraph or None.
        """
        headers = [
            p for p in self.paragraphs
            if isinstance(p, CopyrightHeaderParagraph)]
        if headers:
            return headers[0]

    def merge_contiguous_unknown_paragraphs(self):
        """
        Update self.paragraphs, merging contiguous unknown-only
        CatchAllParagraph paragraphs in one.
        """
        paragraphs = []
        for typ, contigs in itertools.groupby(self.paragraphs, type):
            contigs = list(contigs)
            if typ != CatchAllParagraph or len(contigs) == 1:
                paragraphs.extend(contigs)
                continue

            if not all(p.is_all_unknown() for p in contigs):
                paragraphs.extend(contigs)
                continue

            values = []
            start_line = 1
            end_line = 1
            for para in contigs:
                values.extend(k for k in para.to_dict().values())
                # the new start and end lines are the minimal first line and the
                # maximal last line of contiguous paragraphs
                first, last = para.get_first_last_line_numbers()
                try:
                    start_line = min([start_line, first])
                    end_line = max([end_line, last])
                except Exception as e:
                    raise Exception(repr(e), start_line, first, end_line, last) from e

            paragraphs.append(
                CatchAllParagraph(
                    extra_data={'unknown': debcon.from_formatted_lines(values)},
                    line_numbers_by_field={'unknown':(start_line, end_line,)},
                )
            )

        self.paragraphs = paragraphs

    def fold_contiguous_empty_license_followed_by_unknown(self):
        """
        Update self.paragraphs, such that a CatchAllParagraph paragraph with
        "unknown" as license text is merged into a preceding empty (e.g. without
        any text) CopyrightLicenseParagraph paragraph.
        """
        if len(self.paragraphs) <= 2:
            return

        folded_previous = False
        paragraphs = []
        # iterate on (p1,p2), (p2,p3)....
        for para1, para2 in zip(self.paragraphs, self.paragraphs[1:]):
            if folded_previous:
                folded_previous = False
                continue

            if (isinstance(para1, CopyrightLicenseParagraph)
                and para1.is_empty()
                and isinstance(para2, CatchAllParagraph)
                and para2.is_all_unknown()
            ):
                para1.license.name = ''
                para1.license.text = para2.to_dict().get('unknown', '')

                # The updated CopyrightLicenseParagraph paragraph lines extend
                # from its original start line to the end line of the
                # CatchAllParagraph
                start_line, _end_line = para1.line_numbers_by_field.get('license', (1, 1))
                _start_line, end_line = para2.line_numbers_by_field.get('unknown', (1, 1))
                para1.line_numbers_by_field['license'] = (start_line, end_line,)
                folded_previous = True

            paragraphs.append(para1)

        if not folded_previous:
            paragraphs.append(para2)
        self.paragraphs = paragraphs

    def is_valid(self, strict=False):
        """
        Return True if this is a valid Debian Copyright file.
        If `strict` is True, validate strictly against the spec.
        """
        if not self.paragraphs:
            return False
        has_header = False
        has_files = False
        has_license = False
        has_unknown = False
        first = self.paragraphs[0]
        paragraphs = sorted(self.paragraphs, key=lambda x: repr(type(x)))
        for typ, paras in itertools.groupby(paragraphs, type):
            paras = list(paras)
            if typ == CopyrightHeaderParagraph:
                if not strict:
                    has_header = True

                elif (len(paras) == 1 and paras[0].is_valid(strict)
                      and paras[0] == first):
                    has_header = True

            elif typ == CopyrightFilesParagraph:
                has_files = all(p.is_valid(strict) for p in paras)

            elif typ == CopyrightLicenseParagraph:
                has_license = all(p.is_valid(strict) for p in paras)

            elif typ == CatchAllParagraph:
                has_unknown = all(p.is_valid(strict) for p in paras)

            else:
                # unknown paragraph type
                return False

        valid = has_header and has_files or (has_license and has_files)
        if strict:
            return valid and has_unknown
        else:
            return valid



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# SPDX-License-Identifier: MIT
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
# Author: Peter Odding <peter@peterodding.com>
# Last Change: March 9, 2021
# URL: https://github.com/xolox/python-deb-pkg-tools


from unittest import TestCase

from debian_inspector.version import Version


class DebPkgToolsTestCase(TestCase):

    def test_version_sorting(self):
        # Check version sorting implemented on top of `=' and `<<' comparisons.
        expected_order = ['0.1', '0.5', '1.0', '2.0', '3.0', '1:0.4', '2:0.3']
        assert list(sorted(expected_order)) != expected_order
        result = [str(v) for v in sorted(map(Version.from_string, expected_order))]
        assert expected_order == result

    def test_version_comparison(self):
        assert Version.from_string('1.0') > Version.from_string('0.5')
        assert Version.from_string('1:0.5') > Version.from_string('2.0')
        assert not Version.from_string('0.5') > Version.from_string('2.0')

        assert Version.from_string('0.75') >= Version.from_string('0.5')
        assert Version.from_string('0.50') >= Version.from_string('0.5')
        assert Version.from_string('1:0.5') >= Version.from_string('5.0')
        assert not Version.from_string('0.2') >= Version.from_string('0.5')

        assert Version.from_string('0.5') < Version.from_string('1.0')
        assert Version.from_string('2.0') < Version.from_string('1:0.5')
        assert not Version.from_string('2.0') < Version.from_string('0.5')

        assert Version.from_string('0.5') <= Version.from_string('0.75')
        assert Version.from_string('0.5') <= Version.from_string('0.50')
        assert Version.from_string('5.0') <= Version.from_string('1:0.5')
        assert not Version.from_string('0.5') <= Version.from_string('0.2')

        assert Version.from_string('42') == Version.from_string('42')
        assert Version.from_string('0.5') == Version.from_string('0:0.5')
        assert Version.from_string('0.5') != Version.from_string('1.0')

        assert Version.from_string('1') != Version.from_string('0')
        assert not Version.from_string('0.5') != Version.from_string('0:0.5')

        assert Version.from_string("1.3~rc2") < Version.from_string("1.3")



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#




#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

"""
A parser for deb822 data files as used by Debian control and copyright files,
and several related formats.

For details, see:
- https://www.debian.org/doc/debian-policy/ch-controlfields
- https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
- https://datatracker.ietf.org/doc/rfc2822/

Why yet another Debian 822 parser?

This module exists because no existing parser module supports some must-have
features. Neither the standard email module nor the python-debian library, nor
other existing libraries can do these:

- track the start and end line numbers of each field within each paragraph: we
  need line numbers to trace license detection in copyright files.

- lenient parsing recovering from common errors: this helps handling a larger
  variety of copyright files even if they are not exactly well-formed
"""

import re

import attr

from debian_inspector.debcon import read_text_file


def get_paragraphs_as_field_groups(text):
    """
    Yield lists of Deb822Field for each paragraph in a ``text`` each separated
    by one or more empty lines. Raise Exceptions on errors.
    """
    return get_paragraphs_as_field_groups_from_lines(NumberedLine.lines_from_text(text))


def get_paragraphs_as_field_groups_from_file(location):
    """
    Yield lists of Deb822Field for each paragraph in a control file at
    ``location``. Raise Exceptions on errors.
    """
    if not location:
        return []
    return get_paragraphs_as_field_groups(read_text_file(location))


def get_paragraphs_as_field_groups_from_lines(numbered_lines):
    """
    Yield lists of Deb822Field for each paragraph in a ``numbered_lines`` list
    of NumberedLine. Raise Exceptions on errors.
    """
    numbered_lines = list(numbered_lines)
    last_idx = len(numbered_lines) - 1
    fields_group = []
    current_field = None
    for idx, line in enumerate(numbered_lines):

        # blank line: One or more blank line should terminates paragraph (e.g. a
        # fields_group) and starts a new one. There is one exception
        # though which is when the next line is not a new field declaration.
        # While this is not valid, this is a common mistake in copyright files
        # and we want to recover from this so we treat that empty line as a
        # continuation.
        if line.is_blank():
            # peek one line ahead if next is a header field...
            if (
                current_field
                and idx != last_idx
                and not numbered_lines[idx + 1].is_field_declaration()
                and not numbered_lines[idx + 1].is_blank()
            ):
                current_field.add_continuation_line(line)
                continue

            if fields_group:
                fields_group = clean_fields(fields_group)
                yield fields_group
                fields_group = []
                current_field = None
            else:
                # skip empty lines in between paragraphs
                pass

        # continuation line: append this to the current field
        elif current_field and line.is_field_continuation():
            current_field.add_continuation_line(line)

        # new field declaration line: create a new field
        elif line.is_field_declaration():
            current_field = Deb822Field.from_line(line)
            if not current_field:
                raise Exception(f'Invalid field line: {line}')
            fields_group.append(current_field)

        # an unknown line: we yield the curremt group and then yield this as
        # its own group
        else:
            if fields_group:
                fields_group = clean_fields(fields_group)
                yield fields_group

            # craft a synthetic header with name "unknown"
            yield [Deb822Field(name='unknown', lines=[line])]
            fields_group = []
            current_field = None

    # last header fields group: if any: yield this
    if fields_group:
        fields_group = clean_fields(fields_group)
        yield fields_group


def clean_fields(fields):
    """
    Clean and return a ``fields`` list of Deb822Field.
    """
    for hf in (fields or []):
        hf.rstrip()
    return fields


is_field_declaration = re.compile(r'^[a-z]+[a-z0-9\-]*:.*$', re.IGNORECASE).match
is_field_continuation = re.compile(r'^[ \t]+[\S]+.*$', re.IGNORECASE).match


@attr.s(slots=True)
class NumberedLine:
    """
    A text line that tracks its absolute line number. Numbers start at 1.
    """
    number = attr.ib()
    value = attr.ib()

    def is_blank(self):
        return not self.value.strip()

    def is_field_declaration(self):
        """
        Return True if this is a continuation line.

        For example:

        >>> NumberedLine(1, 'foo').is_field_declaration()
        False
        >>> NumberedLine(1, '').is_field_declaration()
        False
        >>> NumberedLine(1, '   ').is_field_declaration()
        False
        >>> NumberedLine(1, '  Some: bar ').is_field_declaration()
        False
        >>> NumberedLine(1, 'Some:').is_field_declaration()
        True
        >>> NumberedLine(1, 'foo: bar').is_field_declaration()
        True
        >>> NumberedLine(1, 'foo:bar').is_field_declaration()
        True
        >>> NumberedLine(1, 'foo: ').is_field_declaration()
        True
        >>> NumberedLine(1, ' .').is_field_declaration()
        False
        >>> NumberedLine(1, '    .').is_field_declaration()
        False
        """
        return bool(is_field_declaration(self.value))

    def is_field_continuation(self):
        """
        Return True if this is a field declaration line.

        For example:

        >>> NumberedLine(1, 'foo').is_field_continuation()
        False
		>>> NumberedLine(1, '').is_field_continuation()
		False
		>>> NumberedLine(1, '   ').is_field_continuation()
		False
		>>> NumberedLine(1, '	').is_field_continuation()
		False
        >>> NumberedLine(1, '   foo').is_field_continuation()
        True
        >>> NumberedLine(1, ' .').is_field_continuation()
        True
        >>> NumberedLine(1, '	.').is_field_continuation()
        True
        >>> NumberedLine(1, ' "').is_field_continuation()
        True
        >>> NumberedLine(1, ' (').is_field_continuation()
        True
        """
        return bool(is_field_continuation(self.value))

    @classmethod
    def lines_from_text(cls, text):
        """
        Return a list of Line from a ``text``
        """
        return [
            cls(number=number, value=value)
            for number, value in enumerate(text.splitlines(False), 1)
        ]

    def to_dict(self):
        return attr.asdict(self)


@attr.s(slots=True)
class Deb822Field:
    """
    A Deb822Field field with a name and a list of NumberedLines.
    """
    # field name, normalized as stripped and lowercase
    name = attr.ib(default=None)

    lines = attr.ib(default=attr.Factory(list))

    @property
    def text(self):
        return '\n'.join(l.value for l in self.lines)

    @property
    def start_line(self):
        return self.lines[0].number

    @property
    def end_line(self):
        return self.lines[-1].number

    def rstrip(self):
        """
        Remove the last lines of these lines they are all blank or empty.
        Return self.
        """
        lines = self.lines
        while lines:
            if not lines[-1].value.strip():
                lines.pop()
            else:
                break
        return self

    def add_continuation_line(self, line):
        self.lines.append(
            NumberedLine(number=line.number, value=line.value.rstrip())
        )

    @classmethod
    def from_line(cls, line):
        """
        Parse a ``line`` Line object as a "Name: value" declaration line and
        return a Deb822Field object. Return None if this is not a deb822 field
        declaration line.
        """
        if not line or not line.is_field_declaration():
            return

        name, _colon, value = line.value.partition(':')
        name = name.strip().lower()
        if not name:
            return

        # There are some cases where Debian copyright files spells the
        # license field of a paragraph as "licence". We must correct this
        # otherwise the license information will be stored under "licence"
        # instead of "license".
        if name == 'licence':
            name = 'license'

        value = value.strip()
        first_line = NumberedLine(number=line.number, value=value)
        return cls(name=name, lines=[first_line])

    def to_dict(self):
        return attr.asdict(self)



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

from os import path

from test_utils import JsonTester  # NOQA

from debian_inspector import copyright


class TestCopyrightFields(JsonTester):
    test_data_dir = path.join(path.dirname(__file__), 'data')

    def test_CopyrightStatementField(self):
        test = ' 2012-12  MyCom  inc. '
        results = copyright.CopyrightStatementField.from_value(test)
        assert results.year_range == '2012-12'
        assert results.holder == 'MyCom inc.'
        assert results.dumps() == '2012-12 MyCom inc.'

    def test_CopyrightStatementField_no_year(self):
        test = '   MyCom  inc. '
        results = copyright.CopyrightStatementField.from_value(test)
        assert results.holder == 'MyCom inc.'
        assert not results.year_range
        assert results.dumps() == 'MyCom inc.'

    def test_is_year_range(self):
        assert copyright.is_year_range('2012')
        assert copyright.is_year_range('2012-1999')
        assert not copyright.is_year_range('2012-now')
        assert not copyright.is_year_range(' MyCom ')
        assert not copyright.is_year_range(' ')
        assert not copyright.is_year_range('')
        assert not copyright.is_year_range(None)

    def test_LicenseField(self):
        test = '  sim ple   '
        results = copyright.LicenseField.from_value(test)
        assert results.name == 'sim ple'
        assert not results.text
        assert results.dumps() == 'sim ple'

    def test_LicenseField_with_text(self):
        test = ''' GPL 2.0
 licensed under the gpl
 .
  attribution
 .
'''
        results = copyright.LicenseField.from_value(test)
        assert results.name == 'GPL 2.0'
        assert results.text == 'licensed under the gpl\n\n attribution\n'
        assert results.dumps() == 'GPL 2.0\n licensed under the gpl\n .\n  attribution'


class TestDebianCopyright(JsonTester):
    test_data_dir = path.join(path.dirname(__file__), 'data')

    def test_DebianCopyright_from_file__from_copyrights_dep5_1(self):
        test_file = self.get_test_loc('copyright/dep5-b43-fwcutter.copyright')
        expected_loc = 'copyright/dep5-b43-fwcutter.copyright-expected-DebianCopyright.json'
        results = copyright.DebianCopyright.from_file(test_file)
        self.check_json(results.to_dict(with_lines=True), expected_loc, regen=False)

    def test_DebianCopyright_from_file__from_copyrights_dep5_3(self):
        test_file = self.get_test_loc('copyright/dep5-rpm.copyright')
        expected_loc = 'copyright/dep5-rpm.copyright-expected-DebianCopyright.json'
        results = copyright.DebianCopyright.from_file(test_file)
        self.check_json(results.to_dict(with_lines=True), expected_loc, regen=False)

    def test_DebianCopyright_from_file__from_copyrights_dep5_dropbear(self):
        test_file = self.get_test_loc('copyright/dropbear.copyright')
        expected_loc = 'copyright/dropbear.copyright-expected-DebianCopyright.json'
        results = copyright.DebianCopyright.from_file(test_file)
        self.check_json(results.to_dict(with_lines=True), expected_loc, regen=False)

    def test_DebianCopyright_from_file__from_copyrights_with_duplicated_fields(self):
        test_file = self.get_test_loc('copyright/dupe-field.copyright')
        expected_loc = 'copyright/dupe-field.copyright-expected-DebianCopyright.json'
        results = copyright.DebianCopyright.from_file(test_file)
        self.check_json(results.to_dict(with_lines=True), expected_loc, regen=False)

    def test_DebianCopyright_from_file__from_copyrights_dep5_1_dumps(self):
        test_file = self.get_test_loc('copyright/dep5-b43-fwcutter.copyright')
        expected_loc = 'copyright/dep5-b43-fwcutter.copyright-expected.dumps'
        results = copyright.DebianCopyright.from_file(test_file).dumps()
        self.check_file(results, expected_loc, regen=False)

    def test_DebianCopyright_from_file__from_copyrights_dep5_3_dumps(self):
        test_file = self.get_test_loc('copyright/dep5-rpm.copyright')
        expected_loc = 'copyright/dep5-rpm.copyright-expected.dumps'
        results = copyright.DebianCopyright.from_file(test_file).dumps()
        self.check_file(results, expected_loc, regen=False)

    def test_DebianCopyright_from_file__from_copyrights_dep5_dropbear_dumps(self):
        test_file = self.get_test_loc('copyright/dropbear.copyright')
        expected_loc = 'copyright/dropbear.copyright-expected.dumps'
        results = copyright.DebianCopyright.from_file(test_file).dumps()
        self.check_file(results, expected_loc, regen=False)

    def test_DebianCopyright_from_text__from_copyrights_dep5_dropbear_dumps(self):
        test_file = self.get_test_loc('copyright/dropbear.copyright')
        import io
        with io.open(test_file, encoding='utf-8') as td:
            test_data = td.read()
        expected_loc = 'copyright/dropbear.copyright-expected.dumps'
        results = copyright.DebianCopyright.from_text(test_data).dumps()
        self.check_file(results, expected_loc, regen=False)

    def test_DebianCopyright_from_file_split_paragraphs_correctly_multiple_lines(self):
        test_file = self.get_test_loc('copyright/debian-slim-gpgv.copyright')
        expected_loc = 'copyright/debian-slim-gpgv.copyright-expected-DebianCopyright.json'
        results = copyright.DebianCopyright.from_file(test_file)
        self.check_json(results.to_dict(with_lines=True), expected_loc, regen=False)

    def test_DebianCopyright_from_file_licence_to_license(self):
        test_file = self.get_test_loc('copyright/test-licence-license.copyright')
        expected_loc = 'copyright/test-licence-license.copyright-expected-DebianCopyright.json'
        results = copyright.DebianCopyright.from_file(test_file)
        self.check_json(results.to_dict(with_lines=True), expected_loc, regen=True)



class TestCopyright(JsonTester):
    test_data_dir = path.join(path.dirname(__file__), 'data')

    def test_is_machine_readable_copyright(self):
        text = '''format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: b43-fwcutter
Source: http://linuxwireless.org/en/users/Drivers/b43'''
        assert copyright.is_machine_readable_copyright(text)

    def test_is_machine_readable_copyright_ignore_case(self):
        text = '''Format: http://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
Upstream-Name: b43-fwcutter
Source: http://linuxwireless.org/en/users/Drivers/b43'''
        assert copyright.is_machine_readable_copyright(text)

    def test_is_machine_readable_copyright_fasle(self):
        text = '''homepage: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
        '''
        assert not copyright.is_machine_readable_copyright(text)




"""Generate and work with PEP 425 Compatibility Tags.

copied from pip-20.3.1 pip/tests/unit/test_utils_compatibility_tags.py
download_url: https://raw.githubusercontent.com/pypa/pip/20.3.1/tests/unit/test_utils_compatibility_tags.py

Copyright (c) 2008-2020 The pip developers (see AUTHORS.txt file)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from unittest.mock import patch
import sysconfig

import pytest

import utils_pip_compatibility_tags


@pytest.mark.parametrize(
    "version_info, expected",
    [
        ((2,), "2"),
        ((2, 8), "28"),
        ((3,), "3"),
        ((3, 6), "36"),
        # Test a tuple of length 3.
        ((3, 6, 5), "36"),
        # Test a 2-digit minor version.
        ((3, 10), "310"),
    ],
)
def test_version_info_to_nodot(version_info, expected):
    actual = utils_pip_compatibility_tags.version_info_to_nodot(version_info)
    assert actual == expected


class Testcompatibility_tags(object):
    def mock_get_config_var(self, **kwd):
        """
        Patch sysconfig.get_config_var for arbitrary keys.
        """
        get_config_var = sysconfig.get_config_var

        def _mock_get_config_var(var):
            if var in kwd:
                return kwd[var]
            return get_config_var(var)

        return _mock_get_config_var

    def test_no_hyphen_tag(self):
        """
        Test that no tag contains a hyphen.
        """
        import pip._internal.utils.compatibility_tags

        mock_gcf = self.mock_get_config_var(SOABI="cpython-35m-darwin")

        with patch("sysconfig.get_config_var", mock_gcf):
            supported = pip._internal.utils.compatibility_tags.get_supported()

        for tag in supported:
            assert "-" not in tag.interpreter
            assert "-" not in tag.abi
            assert "-" not in tag.platform


class TestManylinux2010Tags(object):
    @pytest.mark.parametrize(
        "manylinux2010,manylinux1",
        [
            ("manylinux2010_x86_64", "manylinux1_x86_64"),
            ("manylinux2010_i686", "manylinux1_i686"),
        ],
    )
    def test_manylinux2010_implies_manylinux1(self, manylinux2010, manylinux1):
        """
        Specifying manylinux2010 implies manylinux1.
        """
        groups = {}
        supported = utils_pip_compatibility_tags.get_supported(platforms=[manylinux2010])
        for tag in supported:
            groups.setdefault((tag.interpreter, tag.abi), []).append(tag.platform)

        for arches in groups.values():
            if arches == ["any"]:
                continue
            assert arches[:2] == [manylinux2010, manylinux1]


class TestManylinux2014Tags(object):
    @pytest.mark.parametrize(
        "manylinuxA,manylinuxB",
        [
            ("manylinux2014_x86_64", ["manylinux2010_x86_64", "manylinux1_x86_64"]),
            ("manylinux2014_i686", ["manylinux2010_i686", "manylinux1_i686"]),
        ],
    )
    def test_manylinuxA_implies_manylinuxB(self, manylinuxA, manylinuxB):
        """
        Specifying manylinux2014 implies manylinux2010/manylinux1.
        """
        groups = {}
        supported = utils_pip_compatibility_tags.get_supported(platforms=[manylinuxA])
        for tag in supported:
            groups.setdefault((tag.interpreter, tag.abi), []).append(tag.platform)

        expected_arches = [manylinuxA]
        expected_arches.extend(manylinuxB)
        for arches in groups.values():
            if arches == ["any"]:
                continue
            assert arches[:3] == expected_arches



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0 AND MIT
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
# Copyright (c) 2018 Peter Odding
# Author: Peter Odding <peter@peterodding.com>

from os import path
import unittest

from debian_inspector import debcon
from debian_inspector import deps
from debian_inspector import package
from debian_inspector import version


class PackageTestCase(unittest.TestCase):

    def test_find_latest_version(self):
        good = ['name_1.0_all.deb', 'name_0.5_all.deb']
        assert 'name_1.0_all.deb' == path.basename(package.find_latest_version(good).original_filename)
        bad = ['one_1.0_all.deb', 'two_0.5_all.deb']
        self.assertRaises(ValueError, package.find_latest_version, bad)

    def test_find_latest_versions(self):
        packages = ['one_1.0_all.deb', 'one_0.5_all.deb', 'two_1.5_all.deb', 'two_0.1_all.deb']
        results = sorted(
            [path.basename(a.original_filename)
             for a in package.find_latest_versions(packages).values()])
        assert results == sorted(['one_1.0_all.deb', 'two_1.5_all.deb'])

    def test_test_DebArch_from_filename(self):
        filename = '/var/cache/apt/archives/python2.7_2.7.3-0ubuntu3.4_amd64.deb'
        deb = package.DebArchive.from_filename(filename)
        assert deb.name == 'python2.7'
        assert str(deb.version) == '2.7.3-0ubuntu3.4'
        assert deb.architecture == 'amd64'
        assert deb.original_filename == filename

    def test_DebArch_from_filename__raise_ValueError_on_errors(self):
        self.assertRaises(ValueError, package.DebArchive.from_filename, 'python2.7_2.7.3-0ubuntu3.4_amd64.not-a-deb')
        self.assertRaises(ValueError, package.DebArchive.from_filename, 'python2.7.deb')


V = version.Version.from_string


class VersionTestCase(unittest.TestCase):
    # Check each individual operator (to make sure the two implementations
    # agree). We use the Version() class for this so that we test both
    # compare_versions() and the Version() wrapper.

    def test_version_comparison_can_sort(self):
        # Check version sorting implemented on top of `=' and `<<' comparisons.
        expected_order = ['0.1', '0.5', '1.0', '2.0', '3.0', '1:0.4', '2:0.3']
        assert expected_order != list(sorted(expected_order))
        assert list(map(str, sorted(map(V, expected_order)))) == expected_order

    def test_version_comparison_superior(self):
        assert V('1.0') > V('0.5')  # usual semantics
        assert V('1:0.5') > V('2.0')  # unusual semantics
        assert not V('0.5') > V('2.0')  # sanity check

    def test_version_comparison_superior_or_equal(self):
        assert V('0.75') >= V('0.5')  # usual semantics
        assert V('0.50') >= V('0.5')  # usual semantics
        assert V('1:0.5') >= V('5.0')  # unusual semantics
        assert not V('0.2') >= V('0.5')  # sanity check

    def test_version_comparison_inferior(self):
        assert V('0.5') < V('1.0')  # usual semantics
        assert V('2.0') < V('1:0.5')  # unusual semantics
        assert not V('2.0') < V('0.5')  # sanity check

    def test_version_comparison_inferior_or_equal(self):
        assert V('0.5') <= V('0.75')  # usual semantics
        assert V('0.5') <= V('0.50')  # usual semantics
        assert V('5.0') <= V('1:0.5')  # unusual semantics
        assert not V('0.5') <= V('0.2')  # sanity check

    def test_version_comparison_equal(self):
        assert V('42')  # usual semantics == V('42')
        assert V('0:0.5')  # unusual semantics == V('0.5')
        assert V('1.0')  # sanity check == not V('0.5')

    def test_version_comparison_not_equal(self):
        assert V('1') != V('0')  # usual semantics
        assert not V('0.5') != V('0:0.5')  # unusual semantics


class DebArchiveTestCase(unittest.TestCase):

    def test_DebArchive_from_filename(self):
        fn = '/var/cache/apt/archives/python2.7_2.7.3-0ubuntu3.4_amd64.deb'
        debarch = package.DebArchive.from_filename(fn)
        expected = package.DebArchive(
            name='python2.7',
            version=version.Version(epoch=0, upstream='2.7.3', revision='0ubuntu3.4'),
            architecture='amd64',
            original_filename=fn)
        assert debarch == expected

    def test_DebArchive_from_filename_udeb(self):
        fn = '/var/cache/apt/archives/python2.7_2.7.3-0ubuntu3.4_amd64.udeb'
        debarch = package.DebArchive.from_filename(fn)
        expected = package.DebArchive(
            name='python2.7',
            version=version.Version(epoch=0, upstream='2.7.3', revision='0ubuntu3.4'),
            architecture='amd64',
            original_filename=fn)
        assert debarch == expected

    def test_CodeArchive_from_filename(self):
        fn = '/var/cache/apt/archives/python2.7_2.7.3-0ubuntu3.4.orig.tar.gz'
        debarch = package.CodeArchive.from_filename(fn)
        expected = package.CodeArchive(
            name='python2.7',
            version=version.Version(epoch=0, upstream='2.7.3', revision='0ubuntu3.4'),
            original_filename=fn)
        assert debarch == expected

    def test_CodeMetadata_from_filename_dsc(self):
        fn = 'base-files_11.1+deb11u8.dsc'
        debarch = package.CodeMetadata.from_filename(fn)
        expected = package.CodeMetadata(
            name='base-files',
            version=version.Version(epoch=0, upstream='11.1+deb11u8', revision='0'),
            original_filename=fn)
        assert debarch == expected

    def test_CodeMetadata_from_filename_copyright(self):
        fn = 'bash_4.1-3+deb6u2_copyright'
        debarch = package.CodeMetadata.from_filename(fn)
        expected = package.CodeMetadata(
            name='bash',
            version=version.Version(epoch=0, upstream='4.1', revision='3+deb6u2'),
            original_filename=fn)
        assert debarch == expected

    def test_CodeArchive_from_filename_supports_tar_gz_bz2_and_xz(self):
        package.CodeArchive.from_filename('python2.7_2.7.3-0ubuntu3.4.orig.tar.gz')
        package.CodeArchive.from_filename('python2.7_2.7.3-0ubuntu3.4.debian.tar.gz')

        package.CodeArchive.from_filename('python2.7_2.7.3-0ubuntu3.4.orig.tar.bz2')
        package.CodeArchive.from_filename('python2.7_2.7.3-0ubuntu3.4.debian.tar.bz2')

        package.CodeArchive.from_filename('python2.7_2.7.3-0ubuntu3.4.orig.tar.xz')
        package.CodeArchive.from_filename('python2.7_2.7.3-0ubuntu3.4.debian.tar.xz')

    def test_CodeArchive_from_filename_raises_ValueError(self):
        with self.assertRaises(ValueError):
            package.CodeArchive.from_filename('python2.7_2.7.3-0ubuntu3.4.orif.tar.gz')


class ControlTestCase(unittest.TestCase):

    def test_parse_control_fields_1(self):
        deb822_package = debcon.Debian822([
            ('Package', 'python-py2deb'),
            ('Depends', 'python-deb-pkg-tools, python-pip, python-pip-accel'),
            ('Installed-Size', '42'),
        ])
        parsed_info = debcon.parse_control_fields(deb822_package)
        expected = {
            'Package': 'python-py2deb',
            'Depends': deps.AndRelationships((
                deps.Relationship(name=u'python-deb-pkg-tools'),
                deps.Relationship(name=u'python-pip'),
                deps.Relationship(name=u'python-pip-accel')
                )),
            'Installed-Size': 42
        }
        assert parsed_info == expected

    def test_parse_control_fields_2(self):
        unparsed_fields = debcon.Debian822.from_string('''
Package: python3.4-minimal
Version: 3.4.0-1+precise1
Architecture: amd64
Installed-Size: 3586
Pre-Depends: libc6 (>= 2.15)
Depends: libpython3.4-minimal (= 3.4.0-1+precise1), libexpat1 (>= 1.95.8), libgcc1 (>= 1:4.1.1), zlib1g (>= 1:1.2.0), foo | bar
Recommends: python3.4
Suggests: binfmt-support
Conflicts: binfmt-support (<< 1.1.2)
''')

        expected = {
            'Architecture': 'amd64',
            'Conflicts': 'binfmt-support (<< 1.1.2)',
            'Depends': 'libpython3.4-minimal (= 3.4.0-1+precise1), libexpat1 (>= 1.95.8), '
                       'libgcc1 (>= 1:4.1.1), zlib1g (>= 1:1.2.0), foo | bar',
            'Installed-Size': '3586',
            'Package': 'python3.4-minimal',
            'Pre-Depends': 'libc6 (>= 2.15)',
            'Recommends': 'python3.4',
            'Suggests': 'binfmt-support',
            'Version': '3.4.0-1+precise1',
        }

        assert unparsed_fields.to_dict(normalize_names=True) == expected

        parsed_fields = debcon.parse_control_fields(unparsed_fields)

        expected = {
            'Architecture': 'amd64',
            'Conflicts': deps.AndRelationships(relationships=(
                deps.VersionedRelationship(name=u'binfmt-support', operator=u'<<', version=u'1.1.2')
            ,)),
            'Depends': deps.AndRelationships(relationships=(
                deps.VersionedRelationship(name=u'libpython3.4-minimal', operator=u'=', version=u'3.4.0-1+precise1'),
                deps.VersionedRelationship(name=u'libexpat1', operator=u'>=', version=u'1.95.8'),
                deps.VersionedRelationship(name=u'libgcc1', operator=u'>=', version=u'1:4.1.1'),
                deps.VersionedRelationship(name=u'zlib1g', operator=u'>=', version=u'1:1.2.0'),
                deps.OrRelationships(relationships=(
                    deps.Relationship(name=u'foo'),
                    deps.Relationship(name=u'bar')
                ,))
            ),),
            'Installed-Size': 3586,
            'Package': 'python3.4-minimal',
            'Pre-Depends': deps.AndRelationships(relationships=(
                deps.VersionedRelationship(name=u'libc6', operator=u'>=', version=u'2.15')
            ,)),
            'Recommends': deps.AndRelationships(relationships=(deps.Relationship(name=u'python3.4'),)),
            'Suggests': deps.AndRelationships(relationships=(deps.Relationship(name=u'binfmt-support'),)),
            'Version': '3.4.0-1+precise1'}

        assert parsed_fields == expected



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#


import json
import shutil
import os.path

from commoncode import testcase


class JsonTester(testcase.FileBasedTesting):
    test_data_dir = os.path.join(os.path.dirname(__file__), 'data')

    def check_json(self, results, expected_loc, regen=False, sort=False):
        """
        Helper to test a results Python native object against an expected JSON
        file at expected_loc.
        """
        expected_loc = self.get_test_loc(expected_loc, must_exist=False)

        if regen:
            regened_exp_loc = self.get_temp_file()
            with open(regened_exp_loc, 'w') as ex:
                json.dump(results, ex, indent=2, separators=(',', ': '))

            expected_dir = os.path.dirname(expected_loc)
            if not os.path.exists(expected_dir):
                os.makedirs(expected_dir)
            shutil.copy(regened_exp_loc, expected_loc)

        with open(expected_loc, 'rb') as ex:
            expected = json.load(ex)
        if sort:
            assert sorted(results) == sorted(expected)
        else:
            assert json.dumps(results, indent=2) == json.dumps(expected, indent=2)

    def check_file(self, results, expected_loc, regen=False, sort=False):
        """
        Helper to test a results text string against an expected file at
        expected_loc.
        """
        expected_loc = self.get_test_loc(expected_loc)

        if regen:
            regened_exp_loc = self.get_temp_file()
            with open(regened_exp_loc, 'w') as ex:
                ex.write(results)

            expected_dir = os.path.dirname(expected_loc)
            if not os.path.exists(expected_dir):
                os.makedirs(expected_dir)
            shutil.copy(regened_exp_loc, expected_loc)

        with open(expected_loc, 'rb') as ex:
            expected = ex.read()
            expected = expected.decode('utf-8')
        if sort:
            assert sorted(results.splitlines()) == sorted(expected.splitlines())
        else:
            assert results == expected



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0 AND MIT
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
# Copyright (c) 2018 Peter Odding
# Author: Peter Odding <peter@peterodding.com>
# URL: https://github.com/xolox/python-deb-pkg-tools
#

import itertools
from os import path

from attr import attrs
from attr import attrib

from debian_inspector.version import Version

"""
Functions to build and inspect Debian binary package archives (``*.deb``,
``*.udeb`` files as well ``.orig`` and ``.debian`` tarballs).
"""


@attrs
class DebArchive(object):
    """
    A .deb binary package archive.
    """
    name = attrib()
    version = attrib()
    architecture = attrib()
    original_filename = attrib(default=None)

    @classmethod
    def from_filename(cls, filename):
        """
        Parse the filename of a Debian binary package archive and return a
        DebArchive instance. Raise ValueError if the `filename` is not valid.
        """
        if isinstance(filename, DebArchive):
            return filename

        name, version, architecture = get_nva(path.basename(filename))
        return cls(
            name=name,
            version=version,
            architecture=architecture,
            original_filename=filename)

    def to_dict(self):
        data = {}
        data['name'] = self.name
        data['version'] = self.version
        data['architecture'] = self.architecture
        data['original_filename'] = self.original_filename
        return data

    def to_tuple(self):
        """
        Return a tuple of name, Version, architecture suitable for sorting.
        This tuple does not contain the original_filename value.
        """
        return tuple(v for v in self.to_dict().values() if v != 'original_filename')


@attrs
class CodeArchive(object):
    """
    A Debian .orig or .debian package archive.
    These are not architecture-specific.
    For instance in ./pool/main/a/apr-util there are files such as:

    - apr-util_1.5.4.orig.tar.bz2 that contains the original upstream code
    - apr-util_1.6.1-4.debian.tar.xz that contains the Debian patches and
      control files
    """
    name = attrib()
    version = attrib()
    original_filename = attrib(default=None)

    @classmethod
    def from_filename(cls, filename):
        """
        Parse the `filename` of a Debian original package archive and return an OrigArchive instance.
        Raise ValueError if the `filename` is not valid.
        """
        if isinstance(filename, CodeArchive):
            return filename

        name, version, _architecture = get_nva(path.basename(filename))
        return cls(
            name=name,
            version=version,
            original_filename=filename
        )

    def to_dict(self):
        data = {}
        data['name'] = self.name
        data['version'] = self.version
        data['original_filename'] = self.original_filename
        return data

    def to_tuple(self):
        """
        Return a tuple of name, Vresion, architecture suitable for sorting.
        This tuple does not contain the original_filename values.
        """
        return tuple(v for v in self.to_dict().values() if v != 'original_filename')


@attrs
class CodeMetadata(CodeArchive):
    """
    A .dsc, copyright or changelog file present in the debian
    package/metadata archive and contains package information
    on the filename or as file contents.

    For example in ./changelogs/main/d/diffutils/ there are
    files such as:
    - diffutils_3.7-5_copyright
    - diffutils_3.7-5_changelog
    And in .pool/main/b/base-files/ there are files such as:
    - base-files_11.1+deb11u8.dsc
    """


def get_nva(filename):
    """
    Return a tuple of (name string, Version object, archictecture string or
    None) parsed from the `filename` of .deb, .udeb, .orig or .debian archive..
    """
    is_known = False
    if filename.endswith(('.deb', '.udeb', '.dsc')):
        basename, _extension = path.splitext(filename)
        is_known = True

    elif filename.endswith(('_changelog', '_copyright')):
        basename, _, _ = filename.rpartition("_")
        is_known = True

    elif filename.endswith(('.tar.gz', '.tar.xz', '.tar.bz2', '.tar.lzma')):
        # A Format: 3.0 archive.
        # Note that we ignore the legacy .diff.gz files for Format: 1.0
        basename, _, _ = filename.rpartition('.tar.')
        # remove the .orig or .debian
        basename, pkgtype = path.splitext(basename)
        if pkgtype in ('.orig', '.debian'):
            is_known = True

    if not is_known:
        raise ValueError(
            'Unknown Debian archive filename format: {}'.format(filename))

    parts = basename.split('_')
    if len(parts) == 2:
        arch = None
        name, evr = parts

    elif len(parts) == 3:
        name, evr , arch = parts

    else:
        raise ValueError(
            'Unknown Debian archive filename format: {}'.format(filename))

    return name, Version.from_string(evr), arch


# TODO: simplify me
def match_relationships(package_archive, relationship_sets):
    """
    Validate that `package_archive` DebArchive satisfies all the relationships
    of a `relationship_sets`. Return True if valid and False otherwise.
    """
    archive_matches = None
    for relationships in relationship_sets:
        status = relationships.matches(package_archive.name, package_archive.version)
        if status is True and archive_matches is not False:
            archive_matches = True
        elif status is False:
            # This package archive specifically conflicts with (at least) one
            # of the given relationship sets.
            archive_matches = False
            # Short circuit the evaluation of further relationship sets because
            # we've already found our answer.
            break
    return archive_matches


def find_latest_version(packages):
    """
    Return the DebArchive package archive with the highest version number given
    a `packages` list of  package filename strings or DebArchive instances.
    All package items must have the same package name

    Raise ValueError if a package filename is invalid or if the packages do not
    have all the same package name.
    """
    if not packages:
        return
    packages = [DebArchive.from_filename(fn) for fn in packages]
    packages = sorted(packages, key=lambda p: p.to_tuple())
    names = set(p.name for p in packages)
    if len(names) > 1:
        msg = 'Cannot compare versions of different package names'
        raise ValueError(msg.format(' '.join(sorted(names))))
    return packages[-1]


def find_latest_versions(packages):
    """
    Return a mapping {name: DebArchive} where DebArchive is the package archive
    with the highest version number for a given package name given a `packages`
    list of  package filename strings or DebArchive instances.
    """
    if not packages:
        return
    packages = sorted([DebArchive.from_filename(fn) for fn in packages])
    latests = {}
    for name, packages_group in itertools.groupby(packages, key=lambda p: p.name):
        latest = find_latest_version(packages_group)
        latests[name] = latest
    return latests



#
# Copyright 2013 Agustin Henze <tin@sluc.org.ar>
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import fnmatch
import os
from os import path

from debian_inspector.copyright import CopyrightFilesParagraph


class Coverage(object):
    """
    Helper to compute the "coverage" of a copyright file "files" paragraph. This
    is used to check if all files in a directory are referenced by a one of the
    patterns in a "files" field.
    """

    # TODO: add support for excludes!!!
    def __init__(self, paragraphs, directory):
        self.paragraphs = paragraphs
        self.directory = directory
        self.unmatched = set()
        self.matched = {}

    def is_perfect(self):
        matched, unmatched = self.compute()
        return matched and not unmatched

    def compute(self):
        """
        Compute the coverage and update self.
        """
        paragraphs = [p for p in self.paragraphs
                      if isinstance(p, CopyrightFilesParagraph)]

        for root, _dirs, files in os.walk(self.directory, topdown=True):
            root = path.relpath(root, self.directory)
            paths = [path.join(root, filename) for filename in files]
            self.unmatch |= set(paths)
            for paragraph in paragraphs:
                for pattern in paragraph.files:
                    goodfiles = []
                    if pattern.find(path.sep) == -1:
                        pattern_norm = path.join(path.curdir, pattern)
                        goodfiles.extend(fnmatch.filter(paths, pattern_norm))
                    goodfiles.extend(fnmatch.filter(paths, pattern))
                    self.unmatched -= set(goodfiles)
                    self.matched.update({f: paragraph for f in goodfiles})

        return self.matched, self.unmatched



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

from collections import defaultdict
import gzip

"""
Utilities to parse a Debian Contents index file.
These are used by apt-file for instance

See https://wiki.debian.org/DebianRepository/Format#A.22Contents.22_indices
for format details.

For example files see:
 - http://ftp.de.debian.org/debian/dists/Debian10.6/main/Contents-amd64.gz
 - http://archive.ubuntu.com/ubuntu/dists/focal/Contents-i386.gz

See also https://salsa.debian.org/debian-irc-team/judd/-/blob/master/supybot/plugins/Judd/debcontents/contents_file.py
"""


def parse_contents(location, has_header=True):
    """
    Return a mapping of {path: [list of packages]} and a mapping of
    {package: [list of paths]} from parsing a Debian Contents file at
    ``location``.
    The Contents file are typically gzipped but we also accept plain text files.

    If ``has_header`` is True, the file is expected to have a header narrative
    and a FILE/LOCATION columns headers before the table starts in earnest.

    See https://wiki.debian.org/DebianRepository/Format#A.22Contents.22_indices
    for format details.
    """
    if location.endswith('.gz'):
        opener, mode = gzip.GzipFile, 'rb'
    else:
        opener, mode = open, 'r'

    packages_by_path = defaultdict(list)
    paths_by_package = defaultdict(list)
    with opener(location, mode=mode) as lines:
        if has_header:
            # keep track if we are now in the table proper
            # e.g. after the FILE  LOCATION header
            # this is the case for Ubuntu
            in_table = False
        else:
            # if we have no header (like in Debian) we start right away.
            in_table = True

        for line in lines:
            if isinstance(line, bytes):
                line = line.decode('utf-8')
            left, _, right = line.strip().rpartition(' ')
            left = left.strip()
            right = right.strip()
            if left == 'FILE' and right == 'LOCATION':
                if not has_header:
                    raise Exception(
                        'Invalid Contents file with a FILE/LOCATION header: '
                        'call with has_header=True.'
                    )

                if not in_table:
                    # The first row of the table SHOULD have the columns "FILE"
                    # and "LOCATION": This is the spec and used to be True for
                    # Debian. But nowadays only Ubuntu older do this.
                    in_table = True
            else:
                if not in_table:
                    continue
                path = left
                packages = right
                package_names = packages.split(',')
                for archsec_name in package_names:
                    # "A list of qualified package names, separated by comma. A
                    # qualified package name has the form
                    # [[$AREA/]$SECTION/]$NAME, where $AREA is the archive area,
                    # $SECTION the package section, and $NAME the name of the
                    # package."

                    # NOTE: we ignore the arch and section for now
                    archsec, _, package_name = archsec_name.rpartition('/')
                    arch, _, section = archsec.rpartition('/')
                    packages_by_path[path].append(package_name)
                    paths_by_package[package_name].append(path)

    if not in_table:
        raise Exception('Invalid Content files without FILE/LOCATION header.')
    return packages_by_path, paths_by_package


if __name__ == '__main__':

    import sys
    import time
    try:
        location = sys.argv[1]
        start = time.time()
        packages_by_path, paths_by_package = parse_contents(location, has_header=False)
        duration = time.time() - start
        print(f'Parsing completed in {duration} seconds.')
        names_count = len(paths_by_package)
        paths_count = len(packages_by_path)
        print(f'Found {names_count} package names with {paths_count} paths.')

    except Exception as e:
        print('Parse a Debian Contents files and print stats.')
        print('Usage: contents <path to a Gzipped Debian Contents index>')
        print('For example, download this file: http://ftp.de.debian.org/debian/dists/Debian10.6/main/Contents-amd64.gz')
        raise




#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0 AND MIT
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
# Copyright (c) 2018 Peter Odding
# Author: Peter Odding <peter@peterodding.com>
# URL: https://github.com/xolox/python-deb-pkg-tools


import re
import unittest

from debian_inspector import deps


class DepsTestCase(unittest.TestCase):

    def test_relationship_parsing(self):
        relationship_set = deps.parse_depends('foo, bar (>= 1) | baz')
        expected = deps.AndRelationships(relationships=(
            deps.Relationship(name='foo'),
            deps.OrRelationships(relationships=(
                deps.VersionedRelationship(name='bar', operator='>=', version='1'),
                deps.Relationship(name='baz')))))
        assert relationship_set == expected

    def test_relationship_parsing_single_relationship(self):
        expected = deps.AndRelationships(relationships=(
            deps.VersionedRelationship(name='foo', operator='=', version='1.0'),))
        assert deps.parse_depends('foo (=1.0)') == expected

    def test_relationship_parsing_raise_valueerror_for_invalid_relationship(self):
        self.assertRaises(ValueError, deps.parse_depends, 'foo (bar) (baz)')
        self.assertRaises(ValueError, deps.parse_depends, 'foo (bar baz qux)')

    def test_parse_depends(self):
        depends = deps.parse_depends('python (>= 2.6), python (<< 3)')
        expected = deps.AndRelationships(relationships=(
            deps.VersionedRelationship(name='python', operator='>=', version='2.6'),
            deps.VersionedRelationship(name='python', operator='<<', version='3'))
        )
        assert depends == expected
        assert not depends.matches('python', '2.5')
        assert depends.matches('python', '2.6')
        assert depends.matches('python', '2.7')

    def test_parse_alternatives_with_no_alternative(self):
        depends = deps.parse_alternatives('python2.6')
        expected = deps.Relationship(name='python2.6')
        assert depends == expected

    def test_parse_alternatives(self):
        depends = deps.parse_alternatives('python2.6 | python2.7')
        expected = deps.OrRelationships(relationships=(
            deps.Relationship(name='python2.6'),
            deps.Relationship(name='python2.7'))
        )
        assert depends == expected

    def test_architecture_restriction_parsing(self):
        relationship_set = deps.parse_depends('qux [i386 amd64]')
        assert 'qux' == relationship_set.relationships[0].name
        assert 2 == len(relationship_set.relationships[0].architectures)
        assert 'i386' in relationship_set.relationships[0].architectures
        assert 'amd64' in relationship_set.relationships[0].architectures

    def test_relationships_objects_as_strings(self):

        def strip(text):
            return re.sub(r'\s+', '', text)

        relationship_set = deps.parse_depends('foo, bar(>=1)|baz[i386]')
        expected = 'foo, bar (>= 1) | baz [i386]'
        assert str(relationship_set) == expected

        expected = deps.AndRelationships(relationships=(
            deps.Relationship(name='foo'),
            deps.OrRelationships(relationships=(
                deps.VersionedRelationship(name='bar', operator='>=', version='1'),
                deps.Relationship(name='baz', architectures=('i386',))))))
        assert relationship_set == expected

    def test_relationship_evaluation_works_without_version(self):
        relationship_set = deps.parse_depends('python')
        assert relationship_set.matches('python')
        assert not relationship_set.matches('python2.7')
        assert ['python'] == list(relationship_set.names)

    def test_relationship_evaluation_alternative_OR_works_without_version(self):
        relationship_set = deps.parse_depends('python2.6 | python2.7')
        assert not relationship_set.matches('python2.5')
        assert relationship_set.matches('python2.6')
        assert relationship_set.matches('python2.7')
        assert not relationship_set.matches('python3.0')
        assert ['python2.6', 'python2.7'] == sorted(relationship_set.names)

    def test_relationship_evaluation_works_without_version_against_versioned(self):
        # Testing for matches without providing a version is valid (should not
        # raise an error) but will never match a relationship with a version.
        relationship_set = deps.parse_depends('python (>= 2.6), python (<< 3)')
        assert relationship_set.matches('python', '2.7')
        assert not relationship_set.matches('python')
        assert ['python'] == list(relationship_set.names)

    def test_relationship_evaluation_combination_AND_works_with_version(self):
        # Distinguishing between packages whose name was matched but whose
        # version didn't match vs packages whose name wasn't matched.
        relationship_set = deps.parse_depends('python (>= 2.6), python (<< 3) | python (>= 3.4)')
        # name matched, version didn't
        assert relationship_set.matches('python', '2.5') is False
        # name didn't match
        assert relationship_set.matches('python2.6') is None
        # name in alternative matched, version didn't
        assert relationship_set.matches('python', '3.0') is False

        # name and version match
        assert relationship_set.matches('python', '2.7') is True
        assert relationship_set.matches('python', '2.6')
        assert relationship_set.matches('python', '3.4')
        assert ['python'] == list(relationship_set.names)

    def test_parse_depends_misc(self):
        dependencies = deps.parse_depends('python (>= 2.6), python (<< 3) | python (>= 3.4)')
        expected = deps.AndRelationships(relationships=(
            deps.VersionedRelationship(name='python', operator='>=', version='2.6'),
            deps.OrRelationships(relationships=(
                deps.VersionedRelationship(name='python', operator='<<', version='3'),
                deps.VersionedRelationship(name='python', operator='>=', version='3.4')
            ,))
        ,))
        assert dependencies == expected

        expected = 'python (>= 2.6), python (<< 3) | python (>= 3.4)'
        assert str(dependencies) == expected

    def test_parse_relationship(self):
        rel = deps.parse_relationship('python')
        assert rel == deps.Relationship(name='python')
        rel = deps.parse_relationship('python (<< 3)')
        assert rel == deps.VersionedRelationship(name='python', operator='<<', version='3')

    def test_parse_depends_complex_cases_is_idempotent(self):
        depends = [
            'clamav-freshclam (>= 0.103.2+dfsg) | clamav-data, libc6 (>= 2.15), libclamav9 (>= 0.103.2), '
                'libcurl3 (>= 7.16.2), libjson-c2 (>= 0.10), libssl1.0.0 (>= 1.0.0), zlib1g (>= 1:1.2.3.3)',
            'sysvinit-utils (>= 2.86.ds1-62), insserv (>> 1.12.0-10)',
            'libgimp2.0 (>= 2.8.16), libgimp2.0 (<= 2.8.16-z)',
            'zlib1g (>= 1:1.1.4), python:any (>= 2.6.6-7~)',
            'git (>> 1:2.34.1), git (<< 1:2.34.1-.)',
            'libnspr4 (>= 2:4.13.1), libnspr4 (<= 2:4.13.1-0ubuntu0.16.04.1.1~)',
            'libc6 (>> 2.23), libc6 (<< 2.24)',
            'libptexenc1 (>= 2015.20160222.37495-1ubuntu0.1), libptexenc1 (<< 2015.20160222.37495-1ubuntu0.1.1~), '
                'libkpathsea6 (>= 2015.20160222.37495-1ubuntu0.1), libkpathsea6 (<< 2015.20160222.37495-1ubuntu0.1.1~), '
                'libsynctex1 (>= 2015.20160222.37495-1ubuntu0.1), libsynctex1 (<< 2015.20160222.37495-1ubuntu0.1.1~), '
                'libtexlua52 (>= 2015.20160222.37495-1ubuntu0.1), libtexlua52 (<< 2015.20160222.37495-1ubuntu0.1.1~), '
                'libtexluajit2 (>= 2015.20160222.37495-1ubuntu0.1), libtexluajit2 (<< 2015.20160222.37495-1ubuntu0.1.1~)',
            'libldb1 (<< 2:1.1.25~), libldb1 (>> 2:1.1.24~)',
            'libnettle6 (= 3.2-1ubuntu0.16.04.2), libhogweed4 (= 3.2-1ubuntu0.16.04.2), libgmp10-dev, dpkg (>= 1.15.4) | install-info',
            'libkf5widgetsaddons-data (= 5.18.0-0ubuntu1), libc6 (>= 2.14), libqt5core5a (>= 5.5.0), '
                'libqt5gui5 (>= 5.3.0) | libqt5gui5-gles (>= 5.3.0), libqt5widgets5 (>= 5.2.0), libstdc++6 (>= 4.1.1)',
            'libnuma1 (= 2.0.11-1ubuntu1.1), libc6-dev | libc-dev',
            'perl, perl (>= 5.11.2) | libextutils-cbuilder-perl, perl (>= 5.12) | libextutils-parsexs-perl, '
                'perl (>= 5.13.9) | libmodule-metadata-perl, perl (>= 5.13.9) | libversion-perl (>= 1:0.8700), '
                'perl (>= 5.19.5) | libtest-harness-perl (>= 3.29), perl (>= 5.21.3) | libcpan-meta-perl (>= 2.142060)',
            'libc6 (>= 2.14), libglib2.0-0 (>= 2.37.3), libgtk-3-0 (>= 3.11.5), libgtksourceview-3.0-1 (>= 3.17.3), '
                'libpeas-1.0-0 (>= 1.0.0), libzeitgeist-2.0-0 (>= 0.9.9), dconf-gsettings-backend | gsettings-backend, '
                'python3 (<< 3.6), python3 (>= 3.5~), python3.5, gedit (>= 3.18), gedit (<< 3.19), gir1.2-git2-glib-1.0, '
                'gir1.2-glib-2.0, gir1.2-gtk-3.0, gir1.2-gtksource-3.0, gir1.2-gucharmap-2.90, gir1.2-pango-1.0, gir1.2-peas-1.0, '
                'gir1.2-vte-2.91, gir1.2-zeitgeist-2.0, python3-gi, python3-gi-cairo, python3-cairo, python3-dbus',
        ]

        for depend in depends:
            assert str(deps.parse_depends(depend)) == depend



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# ScanCode is a trademark of nexB Inc.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/skeleton for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import subprocess
import unittest
import configparser


class BaseTests(unittest.TestCase):
    def test_skeleton_codestyle(self):
        """
        This test shouldn't run in proliferated repositories.
        """
        setup_cfg = configparser.ConfigParser()
        setup_cfg.read("setup.cfg")
        if setup_cfg["metadata"]["name"] != "skeleton":
            return

        args = "venv/bin/black --check -l 100 setup.py etc tests"
        try:
            subprocess.check_output(args.split())
        except subprocess.CalledProcessError as e:
            print("===========================================================")
            print(e.output)
            print("===========================================================")
            raise Exception(
                "Black style check failed; please format the code using:\n"
                "  python -m black -l 100 setup.py etc tests",
                e.output,
            ) from e



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

from os import path

from test_utils import JsonTester  # NOQA

from debian_inspector import deb822
from debian_inspector.deb822 import Deb822Field
from debian_inspector.deb822 import NumberedLine


def get_paras_data(test_file):
    return [
        [h.to_dict() for h in p]
        for p in deb822.get_paragraphs_as_field_groups_from_file(test_file)
    ]


class TestGetParagraphsData(JsonTester):
    test_data_dir = path.join(path.dirname(__file__), 'data')

    def test_get_paragraphs_as_field_groups__splits_paragraphs_with_multiple_lines_correctly(self):
        test = """Upstream-Name: GnuPG

Files: *
Copyright: Free Software Foundation, Inc
License: GPL-3+

License: TinySCHEME
 Redistribution


License: permissive
 This file is free software.

License: RFC-Reference
 doc/OpenPGP


License: GPL-3+
 GnuPG
  . formatted  
   also formatted 
"""
        results = list(deb822.get_paragraphs_as_field_groups(test))
        expected = [
            [Deb822Field(lines=[NumberedLine(number=1, value='GnuPG')], name='upstream-name')],
            [
                Deb822Field(lines=[NumberedLine(number=3, value='*')], name='files'),
                Deb822Field(lines=[NumberedLine(number=4, value='Free Software Foundation, Inc')], name='copyright'),
                Deb822Field(lines=[NumberedLine(number=5, value='GPL-3+')], name='license')
            ],
            [Deb822Field(lines=[
                NumberedLine(number=7, value='TinySCHEME'),
                NumberedLine(number=8, value=' Redistribution')
            ], name='license')],
            [Deb822Field(lines=[
                NumberedLine(number=11, value='permissive'),
                NumberedLine(number=12, value=' This file is free software.')
            ], name='license')],
            [Deb822Field(lines=[
                NumberedLine(number=14, value='RFC-Reference'),
                NumberedLine(number=15, value=' doc/OpenPGP')
            ], name='license')],
            [Deb822Field(name='license', lines=[
                NumberedLine(number=18, value='GPL-3+'), 
                NumberedLine(number=19, value=' GnuPG'), 
                NumberedLine(number=20, value='  . formatted'), 
                NumberedLine(number=21, value='   also formatted')
            ])],
        ]

        assert results == expected

    def test_get_paragraphs_as_field_groups__splits_paragraphs_correctly(self):
        test = 'para1: test1\n\npara2: test2'
        results = list(deb822.get_paragraphs_as_field_groups(test))
        expected = [
            [Deb822Field(lines=[NumberedLine(number=1, value='test1')], name='para1')],
            [Deb822Field(lines=[NumberedLine(number=3, value='test2')], name='para2')],
        ]
        assert results == expected

    def test_get_paragraphs_as_field_groups__handles_more_than_two_empty_lines(self):
        test = 'para1: test1\n\n\n\n\npara2: test2\n test3'
        results = list(deb822.get_paragraphs_as_field_groups(test))
        expected = [
            [Deb822Field(name='para1', lines=[NumberedLine(number=1, value='test1')])],
            [Deb822Field(name='para2', lines=[
                NumberedLine(number=6, value='test2'),
                NumberedLine(number=7, value=' test3')
            ])],
        ]
        assert results == expected

    def test_get_paragraphs_as_field_groups__handles_empty_lines_with_spaces(self):
        test = '\n\npara1: test1\n\n \t     \n          \npara2: test2'
        results = list(deb822.get_paragraphs_as_field_groups(test))
        expected = [
            [Deb822Field(name='para1', lines=[NumberedLine(number=3, value='test1')])],
            [Deb822Field(name='para2', lines=[NumberedLine(number=7, value='test2')])],
        ]
        assert results == expected

    def test_get_paragraphs_as_field_groups_from_file__from_copyrights_dep5_1(self):
        test_file = self.get_test_loc('deb822/dep5-b43-fwcutter.copyright')
        expected_loc = 'deb822/dep5-b43-fwcutter.copyright-expected.json'
        results = get_paras_data(test_file)
        self.check_json(results, expected_loc, regen=False)

    def test_get_paragraphs_as_field_groups_from_file__from_copyrights_dep5_3(self):
        test_file = self.get_test_loc('deb822/dep5-rpm.copyright')
        expected_loc = 'deb822/dep5-rpm.copyright-expected.json'
        results = get_paras_data(test_file)
        self.check_json(results, expected_loc, regen=False)

    def test_get_paragraphs_as_field_groups_from_file__from_copyrights_dep5_dropbear(self):
        test_file = self.get_test_loc('deb822/dropbear.copyright')
        expected_loc = 'deb822/dropbear.copyright-expected.json'
        results = get_paras_data(test_file)
        self.check_json(results, expected_loc, regen=False)

    def test_get_paragraphs_as_field_groups_from_file__from_copyrights_with_spaces(self):
        test_file = self.get_test_loc('deb822/empty_lines.copyright')
        expected_loc = 'deb822/empty_lines.copyright-expected.json'
        results = get_paras_data(test_file)
        self.check_json(results, expected_loc, regen=False)



#!/usr/bin/python

# SPDX-License-Identifier: GPL-2.0-or-later
#

# libdpkg - Debian packaging suite library routines
# t-version.c - test version handling
#
# Copyright © 2009-2014 Guillem Jover <guillem@debian.org>
#
# This is free software you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation either version 2 of the License, or
# (at your option) any later version.
#
# This is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from unittest import TestCase

from debian_inspector import version
from unittest.case import expectedFailure

DPKG_VERSION_OBJECT = version.Version
dpkg_version_compare = version.compare_versions
dpkg_version_relate = version.eval_constraint
parseversion = version.Version.from_string


class VersionTests(TestCase):

    def test_version_compare(self):
        a = DPKG_VERSION_OBJECT(0, "1", "1")
        b = DPKG_VERSION_OBJECT(0, "2", "1")
        assert not dpkg_version_compare(a, b) == 0

        a = DPKG_VERSION_OBJECT(0, "1", "1")
        b = DPKG_VERSION_OBJECT(0, "1", "2")
        assert not dpkg_version_compare(a, b) == 0

        #  Test for version equality.
        a = b = DPKG_VERSION_OBJECT(0, "0", "0")
        assert dpkg_version_compare(a, b) == 0

        a = DPKG_VERSION_OBJECT(0, "0", "00")
        b = DPKG_VERSION_OBJECT(0, "00", "0")
        assert dpkg_version_compare(a, b) == 0

        a = b = DPKG_VERSION_OBJECT(1, "2", "3")
        assert dpkg_version_compare(a, b) == 0

        #  Test for epoch difference.
        a = DPKG_VERSION_OBJECT(0, "0", "0")
        b = DPKG_VERSION_OBJECT(1, "0", "0")
        assert dpkg_version_compare(a, b) < 0
        assert dpkg_version_compare(b, a) > 0

        #  Test for version component difference.
        a = DPKG_VERSION_OBJECT(0, "a", "0")
        b = DPKG_VERSION_OBJECT(0, "b", "0")
        assert dpkg_version_compare(a, b) < 0
        assert dpkg_version_compare(b, a) > 0

        #  Test for revision component difference.
        a = DPKG_VERSION_OBJECT(0, "0", "a")
        b = DPKG_VERSION_OBJECT(0, "0", "b")
        assert dpkg_version_compare(a, b) < 0
        assert dpkg_version_compare(b, a) > 0

        #  FIXME: Complete.

    def test_version_relate(self):

        a = DPKG_VERSION_OBJECT(0, "1", "1")
        b = DPKG_VERSION_OBJECT(0, "1", "1")
        assert dpkg_version_relate(a, '=', b)
        assert not dpkg_version_relate(a, '<<', b)
        assert dpkg_version_relate(a, '<=', b)
        assert not dpkg_version_relate(a, '>>', b)
        assert dpkg_version_relate(a, '>=', b)

        a = DPKG_VERSION_OBJECT(0, "1", "1")
        b = DPKG_VERSION_OBJECT(0, "2", "1")
        assert not dpkg_version_relate(a, '=', b)
        assert dpkg_version_relate(a, '<<', b)
        assert dpkg_version_relate(a, '<=', b)
        assert not dpkg_version_relate(a, '>>', b)
        assert not dpkg_version_relate(a, '>=', b)

        a = DPKG_VERSION_OBJECT(0, "2", "1")
        b = DPKG_VERSION_OBJECT(0, "1", "1")
        assert not dpkg_version_relate(a, '=', b)
        assert not dpkg_version_relate(a, '<<', b)
        assert not dpkg_version_relate(a, '<=', b)
        assert dpkg_version_relate(a, '>>', b)
        assert dpkg_version_relate(a, '>=', b)

    def test_version_parse(self):
        b = DPKG_VERSION_OBJECT(0, "0", "")

        a = parseversion("0")
        assert dpkg_version_compare(a, b) == 0

        a = parseversion("0:0")
        assert dpkg_version_compare(a, b) == 0

        b = DPKG_VERSION_OBJECT(0, "0", "0")
        a = parseversion("0:0-0")
        assert dpkg_version_compare(a, b) == 0

        b = DPKG_VERSION_OBJECT(0, "0.0", "0.0")
        a = parseversion("0:0.0-0.0")
        assert dpkg_version_compare(a, b) == 0

        #  Test epoched versions.
        b = DPKG_VERSION_OBJECT(1, "0", "")
        a = parseversion("1:0")
        assert dpkg_version_compare(a, b) == 0

        b = DPKG_VERSION_OBJECT(5, "1", "")
        a = parseversion("5:1")
        assert dpkg_version_compare(a, b) == 0

        #  Test multiple hyphens.
        b = DPKG_VERSION_OBJECT(0, "0-0", "0")
        a = parseversion("0:0-0-0")
        assert dpkg_version_compare(a, b) == 0

        b = DPKG_VERSION_OBJECT(0, "0-0-0", "0")
        a = parseversion("0:0-0-0-0")
        assert dpkg_version_compare(a, b) == 0

    @expectedFailure
    def test_version_parse_fails_with_multiple_colons_while_dpkg_pass1(self):
        #  Test multiple colons.
        b = DPKG_VERSION_OBJECT(0, "0:0", "0")
        a = parseversion("0:0:0-0")
        assert dpkg_version_compare(a, b) == 0

    @expectedFailure
    def test_version_parse_fails_with_multiple_colons_while_dpkg_pass2(self):
        b = DPKG_VERSION_OBJECT(0, "0:0:0", "0")
        a = parseversion("0:0:0:0-0")
        assert dpkg_version_compare(a, b) == 0

    @expectedFailure
    def test_version_parse_fails_with_multiple_colons_while_dpkg_pass3(self):
        #  Test multiple hyphens and colons.
        b = DPKG_VERSION_OBJECT(0, "0:0-0", "0")
        a = parseversion("0:0:0-0-0")
        assert dpkg_version_compare(a, b) == 0

    @expectedFailure
    def test_version_parse_fails_with_multiple_colons_while_dpkg_pass4(self):
        b = DPKG_VERSION_OBJECT(0, "0-0:0", "0")
        a = parseversion("0:0-0:0-0")
        assert dpkg_version_compare(a, b) == 0

    @expectedFailure
    def test_version_parse_fails_with_multiple_colons_while_dpkg_pass5(self):
        #  Test valid characters in upstream version.
        b = DPKG_VERSION_OBJECT(0, "09azAZ.-+~:", "0")
        a = parseversion("0:09azAZ.-+~:-0")
        assert dpkg_version_compare(a, b) == 0

    @expectedFailure
    def test_version_parse_handles_invalid_chars_in_dpkg1(self):
        #  Test valid characters in revision.
        b = DPKG_VERSION_OBJECT(0, "0", "azAZ09.+~")
        a = parseversion("0:0-azAZ09.+~")
        assert dpkg_version_compare(a, b) == 0

    def test_version_parse_handles_valid_chars_in_revision(self):
        #  Test valid characters in revision.
        b = DPKG_VERSION_OBJECT(0, "0", "azA.+~Z09")
        a = parseversion("0:0-azA.+~Z09")
        assert dpkg_version_compare(a, b) == 0

    def test_version_parse_handles_invalid_chars_in_dpkg2(self):
        #  Test version with leading and trailing spaces.
        b = DPKG_VERSION_OBJECT(0, "0", "1")
        a = parseversion("      0:0-1")
        assert dpkg_version_compare(a, b) == 0
        a = parseversion("0:0-1      ")
        assert dpkg_version_compare(a, b) == 0
        a = parseversion("      0:0-1      ")
        assert dpkg_version_compare(a, b) == 0

    def test_version_parse_raise_exception_on_multiple_colons(self):
        self.assertRaises(ValueError, parseversion, "0:0-0:0-0")
        self.assertRaises(ValueError, parseversion, "0:0:0:0-0")
        self.assertRaises(ValueError, parseversion, "0:0:0-0-0")
        self.assertRaises(ValueError, parseversion, "0:0-0:0-0")
        self.assertRaises(ValueError, parseversion, "0:09azAZ.-+~:-0")
        self.assertRaises(ValueError, parseversion, "0:0-azAZ09.+~")

    def test_version_parse_exceptions(self):

        #  Test empty version.
        self.assertRaises(ValueError, parseversion, "")
        self.assertRaises(ValueError, parseversion, " ")

        #  Test empty upstream version after epoch.
        self.assertRaises(ValueError, parseversion, "0:")

        #  Test empty epoch in version.
        self.assertRaises(ValueError, parseversion, ":1.0")

        #  Test empty revision in version.
        self.assertRaises(ValueError, parseversion, "1.0-")

        #  Test version with embedded spaces.
        self.assertRaises(ValueError, parseversion, "0:0 0-1")

        #  Test version with negative epoch.
        self.assertRaises(ValueError, parseversion, "-1:0-1")

        #  Test invalid characters in epoch.
        self.assertRaises(ValueError, parseversion, "a:0-0")
        self.assertRaises(ValueError, parseversion, "A:0-0")

        #  Test version with huge epoch.
        # self.assertRaises(ValueError, parseversion, "999999999999999999999999:0-1")

        #  Test invalid empty upstream version.
        self.assertRaises(ValueError, parseversion, "-0")
        self.assertRaises(ValueError, parseversion, "0:-0")

        #  Test upstream version not starting with a digit
        self.assertRaises(ValueError, parseversion, "0:abc3-0")

        #  Test invalid characters in upstream version.
        for p in "!#@$%&/|\\<>()[]{},_=*^'":
            verstr = '0:0a' + p + '-0'
            self.assertRaises(ValueError, parseversion, verstr)

        #  Test invalid characters in revision.
        self.assertRaises(ValueError, parseversion, "0:0-0:0")

        for p in "!#@$%&/|\\<>()[]{}:,_=*^'":
            verstr = '0:0-0' + p
            self.assertRaises(ValueError, parseversion, verstr)



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0 AND MIT
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
# Copyright (c) 2018 Peter Odding
# Author: Peter Odding <peter@peterodding.com>
# URL: https://github.com/xolox/python-deb-pkg-tools

from collections.abc import Mapping
from collections.abc import MutableMapping
from collections.abc import Sequence
import email
import io
import re
import textwrap

from attr import attrs
from attr import attrib
import chardet

from debian_inspector import unsign

"""
Utilities to parse Debian-style control files aka. deb822 format.

See https://salsa.debian.org/dpkg-team/dpkg/blob/0c9dc4493715ff3b37262528055943c52fdfb99c/man/deb822.man

https://www.debian.org/doc/debian-policy/ch-controlfields#s-f-Description

This is an alternative to a subset of python-debian library with these
characteristics:

 - We use a lenient parsing accepting things that would not be considered
   strictly Debian-compliant.
 - The focus is first on reading Debian files and writing them second.
 - Copyright, package, control and status files are the most interesting formats.
   Changelog and other Debian file types are mostly ignored for now.
 - There is no attention paid to compatibility and support for older formats
   and older Python versions before 3.6.
"""


@attrs
class FieldMixin(object):
    """
    Base mixin for attrs-based fields.
    """

    @classmethod
    def attrib(cls, **kwargs):
        """
        Return an attrib class
        """
        return attrib(converter=cls.from_value, **kwargs)

    @classmethod
    def from_value(cls, value):
        if isinstance(value, cls):
            return value
        return cls(value)

    def dumps(self):
        """
        Return a string in Debian format for this field.
        """
        return NotImplementedError

    def __str__(self, *args, **kwargs):
        return self.dumps()


@attrs
class SingleLineField(FieldMixin):
    """
    https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/#single-line
    """
    value = attrib()

    @classmethod
    def from_value(cls, value):
        return cls(value=value and value.strip())

    def dumps(self):
        return self.value or ''


@attrs
class LineSeparatedField(FieldMixin):
    """
    https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/#line-based-lists
    """
    values = attrib()

    @classmethod
    def from_value(cls, value):
        values = []
        if value:
            for val in line_separated(value):
                values.append(val.strip())
        return cls(values=values)

    def dumps(self, **kwargs):
        return '\n '.join(self.values or [])


@attrs
class LineAndSpaceSeparatedField(FieldMixin):
    """
    This is a list of values where each item is itself a space-separated list.
    """
    values = attrib()

    @classmethod
    def from_value(cls, value):
        values = []
        if value:
            for val in line_separated(value):
                values.append(tuple(space_separated(val)))
        return cls(values=values)

    def dumps(self, **kwargs):
        return '\n '.join(' '.join(v) for v in self.values or [])


@attrs
class AnyWhiteSpaceSeparatedField(FieldMixin):
    """
    https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/#white-space-lists
    This is a list of values where each item is itself a space-separated list.
    """
    values = attrib()

    @classmethod
    def from_value(cls, value):
        values = []
        if value:
            values = [val for val in value.split()]
        return cls(values=values)

    def dumps(self, **kwargs):
        return '\n '.join(self.values or [])


@attrs
class FormattedTextField(FieldMixin):
    """
    https://www.debian.org/doc/debian-policy/ch-controlfields#description
    Like Description, but there is no special meaning for the first line.
    """
    text = attrib()

    @classmethod
    def from_value(cls, value):
        if value:
            value = from_formatted_text(value)
        return cls(text=value)

    def dumps(self, **kwargs):
        lines = line_separated(self.text)
        if not lines:
            return ''
        return as_formatted_lines(lines)


def as_formatted_lines(lines):
    """
    Return a text formatted for use in a Debian control file with proper
    continuation for multilines.
    """
    if not lines:
        return ''
    formatted = []
    for line in lines:
        is_blank = not line.strip()
        if is_blank:
            formatted.append('.')
        else:
            formatted.append(f'{line}')
    return '\n '.join(formatted)


def as_formatted_text(text):
    """
    Return a text formatted for use in a Debian control file with proper
    continuation for multilines.
    """
    if not text:
        return text
    lines = text.splitlines(False)
    return as_formatted_lines(lines)


def from_formatted_text(text):
    """
    Return cleaned text from a Debian formatted description text
    using rules for handling line prefixes and continuations.
    """
    if not text:
        return text
    return from_formatted_lines(line_separated(text))


def from_formatted_lines(lines):
    """
    Return text from a list of `lines` strings using the Debian
    Description rules for handling line prefixes and continuations.
    """
    if not lines:
        return lines

    # first line is always "stripped"
    text = [lines[0].strip()]
    for line in lines[1:]:
        line = line.rstrip()
        if line.startswith('  '):
            # starting with two or more spaces: displayed verbatim.
            text.append(line[1:])
        elif line == (' .'):
            # containing a single space followed by a single full stop
            # character: rendered as blank lines.
            text.append('')
        elif line.startswith(' .'):
            # containing a space, a full stop and some more characters:  for
            # future expansion.... but we keep them for now
            text.append(line[2:])
        elif line.startswith(' '):
            # starting with a single space. kept stripped
            text.append(line.strip())
        else:
            # this should never happen!!!
            # but we keep it too
            text.append(line.strip())
    return '\n'.join(text)


@attrs
class DescriptionField(FieldMixin):
    """
    https://www.debian.org/doc/debian-policy/ch-controlfields#description
    5.6.13. Description
    """
    synopsis = attrib(default=None)
    text = attrib(default=None)

    @classmethod
    def from_value(cls, value):
        value = value or ''
        lines = line_separated(value)
        if lines:
            synopsis = lines[0].strip()
            text = from_formatted_lines(lines[1:])
            return cls(synopsis=synopsis, text=text)
        else:
            return cls(synopsis='')

    def dumps(self, **kwargs):
        """
        Return a string representation of self.
        """
        syn = self.synopsis or ''
        syn = syn.strip()
        dumped = [syn]
        text = self.text or ''
        if text:
            if text.startswith(' '):
                text = text[1:]
            dumped.append(as_formatted_text(text))
        return '\n '.join(dumped)


@attrs
class File(object):
    name = attrib(default=None)
    size = attrib(default=None)
    md5 = attrib(default=None)
    sha1 = attrib(default=None)
    sha256 = attrib(default=None)
    sha512 = attrib(default=None)


@attrs
class FileField(FieldMixin):
    name = attrib(default=None)
    size = attrib(default=None)
    checksum = attrib(default=None)

    @classmethod
    def from_value(cls, value):
        checksum = size = name = None
        if value:
            checksum, size , name = space_separated(value)
        return cls(checksum=checksum, size=size , name=name)

    def dumps(self, **kwargs):
        return '{} {} {}'.format(self.checksum, self.size , self.name)


@attrs
class FilesField(FieldMixin):
    """
    This is a list of FileField
    """
    values = attrib()

    @classmethod
    def from_value(cls, value):
        values = []
        if value:
            for val in line_separated(value):
                values.append(FileField.from_value(val))
        return cls(values=values)

    def dumps(self, **kwargs):
        return '\n '.join(v.dumps(**kwargs) for v in self.values or [])


def collect_files(data):
    """
    Return a mapping of {name: File} from a Debian data mapping.

    Note: the Files and Checksums-* fields have the same structure and
    contain redundant data.
    """
    files = {}
    for name, size, md5 in collect_file(data.get('files', [])):
        f = File(md5, size , name)
        files[name] = f

    for name, size, sha1 in collect_file(data.get('checksums-sha1', [])):
        f = files[name]
        assert f.size == size
        f.sha1 = sha1

    for name, size, sha256 in collect_file(data.get('checksums-sha256', [])):
        f = files[name]
        assert f.size == size
        f.sha256 = sha256

    for name, size, sha512 in collect_file(data.get('checksums-v', [])):
        f = files[name]
        assert f.size == size
        f.sha512 = sha512

    return files


def collect_file(value):
    """
    Yield tuples of (name, size, digest) given a Debian "Files" value string
    which contains digest, size and name.
    """
    for line in line_separated(value):
        digest, size , name = space_separated(line)
        yield name, size, digest


@attrs
class MaintainerField(FieldMixin):
    """
    https://www.debian.org/doc/debian-policy/ch-controlfields#s-f-maintainer
    5.6.2. Maintainer
    """
    name = attrib()
    email_address = attrib(default=None)

    @classmethod
    def from_value(cls, value):
        name = email_address = None
        if value:
            value = value.strip()
            name, email_address = email.utils.parseaddr(value)  # NOQA
            if not name:
                name = value
                email_address = None
            return cls(name=name, email_address=email_address)

    def dumps(self, **kwargs):
        name = self.name
        if self.email_address:
            name = '{} <{}>'.format(name, self.email_address)
        return name.strip()


def get_paragraphs_data_from_file(location):
    """
    Yield paragraph data mappings from the Debian control file at `location`
    that contains multiple paragraphs (e.g. Package, status, copyright file,
    etc.).
    """
    if not location:
        return []
    return get_paragraphs_data(read_text_file(location))


def split_in_paragraphs(text):
    """
    Yield paragraphs from a `text` string that contains one or more paragraph
    separated by empty lines. Each paragraph is a string.
    """
    for p in re.split(r'\n\n(?:[ \t]*\n)*', text or ''):
        if p:
            yield p


def get_paragraphs_data(text):
    """
    Yield paragraph data mappings from the Debian control `text` string that
    contains multiple paragraphs (e.g. Package, status, copyright file, etc.).
    """
    for para in split_in_paragraphs(text or ''):
        yield get_paragraph_data(para)


def get_paragraph_data_from_file(location, remove_pgp_signature=False):
    """
    Return paragraph data from the Debian control file at `location` that
    contains a single paragraph (e.g. a dsc file).

    Optionally remove a wrapping PGP signature if `remove_pgp_signature` is
    True.
    """
    if not location:
        return []
    return get_paragraph_data(
        read_text_file(location),
        remove_pgp_signature=remove_pgp_signature,
    )


def get_paragraph_data(text, remove_pgp_signature=False):
    """
    Return paragraph data from the Debian control `text`. The paragraph data is
    an ordered mapping of {name: value} fields. If there is data that is not
    parsable or not attached to a field name, this will be added to a field
    named "unknown".

    If there are duplicates field names, the string values of duplicates field
    names are merged together with a new line in the first occurence of that
    field.

    Optionally remove a wrapping PGP signature if `remove_pgp_signature` is
    True.
    """
    if not text:
        return {'unknown': text}

    if remove_pgp_signature:
        text = unsign.remove_signature(text)

    try:
        mls = email.message_from_string(text)
    except UnicodeEncodeError:
        t = text.encode('utf-8')
        mls = email.message_from_string(t)

    items = list(mls.items())

    if not items or mls.defects:
        return {'unknown': text}

    # in a header-only email we should not have a payload. Yet when this happens
    # we should no ignore it either, so let's treat this as "unknown"
    payload = mls.get_payload()
    if payload:
        items.append(('unknown', payload))

    data = {}
    for name, value in items:
        # we do not preserve case: debian field names are case-insensitive AND
        # we use a normalized lowercase version throughout.
        name = name.lower().strip()
        value = value.strip()
        if name in data:
            existing_values = data.get(name, '').splitlines()
            if value not in existing_values:
                value = '\n'.join(existing_values + [value])
        data[name] = value

    return data


def line_separated(value):
    """
    Return a list of values from a `value` string using line as list delimiters.
    """
    if not value:
        return []
    return list(value.splitlines(False))


def _splitter(value, separator):
    """
    Return a list of values from a `value` string using `separator` as list delimiters.
    Empty values are NOT returned.
    """
    if not value:
        return []
    return [v.strip() for v in value.split(separator) if v.strip()]


def comma_separated(value):
    return _splitter(value, ',')


def comma_space_separated(value):
    return _splitter(value, ', ')


def space_separated(value):
    """
    Return a list of values from a `value` string using one or more whitespace
    as list items delimiter. Empty values are NOT returned.
    """
    if not value:
        return []
    return list(value.split())


def read_text_file(location):
    """
    Return the content of the file at `location` as text or None.
    """
    if not location:
        return
    try:
        with io.open(location, 'r', encoding='utf-8') as tc:
            return tc.read()
    except UnicodeDecodeError:
        with open(location, 'rb') as tc:
            content = tc.read()
        enc = chardet.detect(content)['encoding']
        return content.decode(enc)


class Debian822(MutableMapping):
    """
    A mapping-like class that corresponds to a single deb822 paragraph like a
    whole .dsc file.
    """

    def __init__(self, data=None):
        """
        Build a new instance from `data` that is either a file-like object with
        a read() method, a text, a sequence of (key/values) or a mapping. Note
        that the keys are always lowercased.
        """
        if data:
            text = None
            if isinstance(data, Mapping):
                paragraph = {k.lower(): v for k, v in data.items()}

            elif isinstance(data, str):
                text = data

            elif hasattr(data, 'read'):
                text = data.read()

            elif isinstance(data, Sequence):
                # a sequence should be a sequence of items or sequence of string
                # (before the : split)
                seq = list(data)
                first = seq[0]
                if isinstance(first, str):
                    seq = (s.partition(': ') for s in seq)
                    paragraph = {k.lower(): v for k, _, v in seq}
                else:
                    # seq of (k, v) items
                    paragraph = {k.lower(): v for k, v in data}

            else:
                raise TypeError(
                    'Invalid argument type. Should be one of a file-like object, '
                    'a text string, a sequence of items or a mapping but is '
                    'instead:'.format(type(data)))
            if text:
                # we parse in a sequence of items
                paragraph = get_paragraph_data(text, remove_pgp_signature=True)

            self.data = paragraph
        else:
            self.data = {}

    def __getitem__(self, key):
        return self.data.__getitem__(key.lower())

    def __setitem__(self, key, value):
        self.data.__setitem__(key.lower(), value)

    def __delitem__(self, key):
        return self.data.__delitem__(key.lower())

    def __iter__(self):
        return self.data.__iter__()

    def __len__(self):
        return self.data.__len__()

    @classmethod
    def from_file(cls, location, remove_pgp_signature=True):
        data = get_paragraph_data_from_file(
            location=location, remove_pgp_signature=remove_pgp_signature)
        if not data:
            raise ValueError('Location has no parsable data: {}'.format(location))
        return Debian822(data)

    @classmethod
    def from_string(cls, text):
        return Debian822(textwrap.dedent(text).strip())

    def to_dict(self, normalize_names=False):
        if normalize_names:
            return {normalize_control_field_name(key): value
                for key, value in self.data.items()}
        else:
            return dict(self.data)

    def __repr__(self):
        return self.dumps()

    def dumps(self, **kwargs):
        """
        Return a text that resembles the original Debian822 format. This is not
        meant to be a high fidelity rendering and not meant to be used as-is in
        control files.
        """
        items = self.items()

        lines = []
        for key, value in items:
            key = normalize_control_field_name(key)
            lines.append('{}: {}'.format(key, value))
        text = '\n'.join(lines) + '\n'
        return text

    def dump(self, file_like=None, **kwargs):
        text = self.dumps(**kwargs)
        if file_like:
            file_like.write(text.encode('utf-8'))
        else:
            return text


DEFAULT_CONTROL_FIELDS = {
    'Architecture': 'all',
    'Priority': 'optional',
    'Section': 'misc',
}


def load_control_file(control_file):
    """
    Load a control file and return the parsed control fields.

    :param control_file: The filename of the control file to load (a string).
    :returns: A dictionary created by :func:`parse_control_fields()`.
    """
    with open(control_file) as inp:
        return parse_control_fields(Debian822(inp))


DEPS_FIELDS = frozenset([
    # Binary control file fields.
    'Breaks',
    'Conflicts',
    'Depends',
    'Enhances',
    'Pre-Depends',
    'Provides',
    'Recommends',
    'Replaces',
    'Suggests',

    # Source control file fields.
    'Build-Conflicts',
    'Build-Conflicts-Arch',
    'Build-Conflicts-Indep',
    'Build-Depends',
    'Build-Depends-Arch',
    'Build-Depends-Indep',
    'Built-Using',
])


def parse_control_fields(input_fields, deps_fields=DEPS_FIELDS):
    """
    Return an ordered mapping from parsing an`input_fields` mapping of Debian
    control file fields. This applies a few conversions such as:

    - The values of the fields that contain dependencies are parsed
      into Python data structures.

    - The value of some fields such as `Installed-Size` from a string to a
      native type (here an integer).
    """
    from debian_inspector import deps
    output_fields = {}
    for name, unparsed_value in input_fields.items():
        name = normalize_control_field_name(name)
        if name in deps_fields:
            parsed_value = deps.parse_depends(unparsed_value)
        elif name == 'Installed-Size':
            parsed_value = int(unparsed_value)
        else:
            parsed_value = unparsed_value
        output_fields[name] = parsed_value
    return output_fields


def normalize_control_field_name(name):
    """
    Return a case-normalized field name string.

    Normalization of control file field names is not really needed when reading
    as we lowercase everything and replace dash to underscore internally, but it
    can help to compare the parsing results to the original file while testing.

    According to the Debian Policy Manual field names are not case-sensitive,
    however a conventional capitalization is most common and not using it may
    break hings.

    http://www.debian.org/doc/debian-policy/ch-controlfields.html#s-controlsyntax
    """
    special_cases = dict(md5sum='MD5sum', sha1='SHA1', sha256='SHA256')
    return '-'.join(special_cases.get(
        w.lower(), w.capitalize())
        for w in name.split('-')
    )



#
# Copyright (c) 2014-2019 Security Innovation, Inc
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0 AND BSD-3-Clause
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import re

"""
Utility to remove PGP signature from messages used to get the RFC822 message
in a .dsc file that uses signature.
Copied from # https://github.com/SecurityInnovation/PGPy/blob/e2893da8b2f5ce0694257caddd8d852aece546ff/pgpy/types.py#L57
and adapted for debian_inspector.
"""


def is_signed(text):
    """
    Return True if the text is likely PGP-signed.
    """
    if text and isinstance(text, str):
        text = text.strip()
        return text and (text.startswith('-----BEGIN PGP SIGNED MESSAGE-----')
                    and text.endswith('-----END PGP SIGNATURE-----'))
    return False


def remove_signature(text):
    """
    Return `text` stripped from a PGP signature if there is one. Return the
    `text` as-is otherwise.
    """
    if not is_signed(text):
        return text

    signed = pgp_signed(text)
    if not signed:
        return text
    unsigned = signed.groupdict().get('cleartext')
    return unsigned

# A re.VERBOSE regular expression to parse a PGP signed message in its parts.
# the re.VERBOSE flag allows for:
#  - whitespace is ignored except when in a character class or escaped
#  - anything after a '#' that is not escaped or in a character class is
#  ignored, allowing for comments


pgp_signed = re.compile(r"""
    # This capture group is optional because it will only be present in signed
    # cleartext messages

    (^-{5}BEGIN\ PGP\ SIGNED\ MESSAGE-{5}(?:\r?\n)
       (Hash:\ (?P<hashes>[A-Za-z0-9\-,]+)(?:\r?\n){2})?
       (?P<cleartext>(.*\r?\n)*(.*(?=\r?\n-{5})))(?:\r?\n)
    )?

    # Armor header line: capture the variable part of the magic text

    ^-{5}BEGIN\ PGP\ (?P<magic>[A-Z0-9 ,]+)-{5}(?:\r?\n)

    # Try to capture all the headers into one capture group.
    # If this doesn't match, m['headers'] will be None

    (?P<headers>(^.+:\ .+(?:\r?\n))+)?(?:\r?\n)?

    # capture all lines of the body, up to 76 characters long, including the
    # newline, and the pad character(s)

    (?P<body>([A-Za-z0-9+/]{1,76}={,2}(?:\r?\n))+)

    # capture the armored CRC24 value

    ^=(?P<crc>[A-Za-z0-9+/]{4})(?:\r?\n)

    # finally, capture the armor tail line, which must match the armor header
    # line

    ^-{5}END\ PGP\ (?P=magic)-{5}(?:\r?\n)?

     """, flags=re.MULTILINE | re.VERBOSE).search  # NOQA



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here.

import pathlib
import sys

srcdir = pathlib.Path(__file__).resolve().parents[2].joinpath('src')
sys.path.insert(0, srcdir.as_posix())


# -- Project information -----------------------------------------------------

project = "debian_inspector"
copyright = "nexB Inc. and others."
author = "AboutCode.org authors and contributors"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinxcontrib.apidoc",
    "sphinx_reredirects",
    "sphinx_rtd_theme",
    "sphinx_rtd_dark_mode",
    "sphinx.ext.extlinks",
    "sphinx_copybutton"
]


# Setting for sphinxcontrib.apidoc to automatically create API documentation.
apidoc_module_dir = srcdir.joinpath('debian_inspector').as_posix()
apidoc_separate_modules=True
apidoc_module_first = True


# Redirects for olds pages
# See https://documatt.gitlab.io/sphinx-reredirects/usage.html
redirects = {}

# This points to aboutcode.readthedocs.io
# In case of "undefined label" ERRORS check docs on intersphinx to troubleshoot
# Link was created at commit - https://github.com/nexB/aboutcode/commit/faea9fcf3248f8f198844fe34d43833224ac4a83

# Reference to other Sphinx documentations
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "aboutcode": ("https://aboutcode.readthedocs.io/en/latest/", None),
    "scancode-workbench": (
        "https://scancode-workbench.readthedocs.io/en/develop/",
        None,
    ),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

master_doc = "index"

html_context = {
    "display_github": True,
    "github_user": "nexB",
    "github_repo": "debian-inspector",
    "github_version": "main",  # branch
    "conf_py_path": "/docs/source/",  # path in the checkout to the docs root
}

html_css_files = [
    "theme_overrides.css",
]


# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = True

# Define CSS and HTML abbreviations used in .rst files.  These are examples.
# .. role:: is used to refer to styles defined in _static/theme_overrides.css and is used like this: :red:`text`
rst_prolog = """
.. |psf| replace:: Python Software Foundation

.. # define a hard line break for HTML
.. |br| raw:: html

   <br />

.. role:: red

.. role:: img-title

.. role:: img-title-para

"""

# -- Options for LaTeX output -------------------------------------------------

latex_elements = {"classoptions": ",openany,oneside"}



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#


from os import path

from test_utils import JsonTester  # NOQA

from debian_inspector import contents


class TestContentsParse(JsonTester):
    test_data_dir = path.join(path.dirname(__file__), 'data')

    def test_parse_contents_ubuntu_with_header_plain(self):
        test_file = self.get_test_loc('contents/ubuntu_Contents-i386')
        expected_loc = 'contents/ubuntu_Contents-i386-expected.json'
        results = contents.parse_contents(test_file, has_header=True)
        self.check_json(results, expected_loc, regen=False)

    def test_parse_contents_debian_no_header_gzipped(self):
        test_file = self.get_test_loc('contents/debian_Contents-amd64.gz')
        expected_loc = 'contents/debian_Contents-amd64.gz-expected.json'
        results = contents.parse_contents(test_file, has_header=False)
        self.check_json(results, expected_loc, regen=False)

    def test_parse_contents_debian_is_same_gzipped_or_not(self):
        test_file = self.get_test_loc('contents/debian_Contents-amd64.gz')
        results = contents.parse_contents(test_file, has_header=False)

        test_file2 = self.get_test_loc('contents/debian_Contents-amd64')
        results2 = contents.parse_contents(test_file2, has_header=False)

        assert results == results2



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0 AND MIT
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
# Copyright (c) 2018 Peter Odding
# Author: Peter Odding <peter@peterodding.com>
# URL: https://github.com/xolox/python-deb-pkg-tools

import os


def find_debian_architecture():
    """
    Find the Debian architecture of the current environment.

    Uses :func:`os.uname()` to determine the current machine architecture
    (the fifth value returned by :func:`os.uname()`) and translates it into
    one of the `machine architecture labels`_ used in the Debian packaging
    system:

    ====================  ===================
    Machine architecture  Debian architecture
    ====================  ===================
    ``i686``              ``i386``
    ``x86_64``            ``amd64``
    ``armv6l``            ``armhf``
    ====================  ===================

    When the machine architecture is not listed above, this function falls back
    to the external command ``dpkg-architecture -qDEB_BUILD_ARCH`` (provided by
    the ``dpkg-dev`` package). This command is not used by default because:

    1. deb-pkg-tools doesn't have a strict dependency on ``dpkg-dev``.
    2. The ``dpkg-architecture`` program enables callers to set the current
       architecture and the exact semantics of this are unclear to me at the
       time of writing (it can't automagically provide a cross compilation
       environment, so what exactly does it do?).

    :returns: The Debian architecture (a string like ``i386``, ``amd64``,
              ``armhf``, etc).
    :raises: :exc:`~executor.ExternalCommandFailed` when the
             ``dpkg-architecture`` program is not available or reports an
             error.

    .. _machine architecture labels: https://www.debian.org/doc/debian-policy/ch-controlfields.html#s-f-Architecture
    .. _more architectures: https://www.debian.org/ports/index.en.html#portlist-released
    """
    _sysname, _nodename, _release, _version, machine = os.uname()
    if machine == 'i686':
        return 'i386'
    elif machine == 'x86_64':
        return 'amd64'
    elif machine == 'armv6l':
        return 'armhf'
    else:
        raise Exception('unknown machine')



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: MIT
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
# Copyright (c) Peter Odding
# Author: Peter Odding <peter@peterodding.com>
# URL: https://github.com/xolox/python-deb-pkg-tools
#

import logging
import operator as operator_module
import re
from functools import cmp_to_key
from itertools import zip_longest

from attr import asdict
from attr import attrs
from attr import attrib

logger = logging.getLogger(__name__)

"""
Parse, compare and sort Debian package versions.

This module is an implementation of the version comparison and sorting algorithm
described at
https://www.debian.org/doc/debian-policy/ch-controlfields.html#s-f-Version

This has been substantially modified and enhanced from the original python-dpkg
Dpkg class by Nathan J. Meh and team from The Climate Corporation and
code from python-deb-pkg-tools by Peter Odding to extract only the subset that
does the version parsing, comparison and version constraints evaluatiob.


So much so that little of this code may still looks like the original.

Some examples:

#### Compare two arbitrary version strings

    >>> from debian_inspector import debver
    >>> debver.compare_versions('0:1.0-test1', '0:1.0-test2')
    -1
    >>> debver.compare_versions('1.0', '0.6')
    1
    >>> debver.compare_versions('2:1.0', '1:1.0')
    -1

#### Use Version as a key function to sort a list of version strings

    >>> from debian_inspector.debver import Version
    >>> sorted(['0:1.0-test1', '1:0.0-test0', '0:1.0-test2'] , key=Version.from_string)
    ['0:1.0-test1', '0:1.0-test2', '1:0.0-test0']

"""


@attrs(
    eq=False,
    order=False,
    frozen=True,
    hash=False,
    slots=True,
    str=False
)
class Version(object):
    """
    Rich comparison of Debian package versions as first-class Python objects.

    The :class:`Version` class is a subclass of the built in :class:`str` type
    that implements rich comparison according to the version sorting order
    defined in the Debian Policy Manual. Use it to sort Debian package versions
    from oldest to newest in ascending version order like this:

      >>> from debian_inspector.version import Version
      >>> unsorted = ['0.1', '0.5', '1.0', '2.0', '3.0', '1:0.4', '2:0.3']
      >>> print([str(v) for v in sorted(Version.from_string(s) for s in unsorted)])
      ['0.1', '0.5', '1.0', '2.0', '3.0', '1:0.4', '2:0.3']

    This example uses 'epoch' numbers (the numbers before the colons) to
    demonstrate that this version sorting order is different from regular
    sorting and 'natural order sorting'.
    """
    epoch = attrib(default=0)
    upstream = attrib(default=None)
    revision = attrib(default='0')

    def __str__(self, *args, **kwargs):
        if self.epoch:
            version = f'{self.epoch}:{self.upstream}'
        else:
            version = f'{self.upstream}'

        if self.revision not in (None, '0'):
            version += f'-{self.revision}'

        return version

    def __repr__(self, *args, **kwargs):
        return str(self)

    def __hash__(self):
        return hash(self.tuple())

    def __eq__(self, other):
        return type(self) is type(other) and self.tuple() == other.tuple()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if type(self) is type(other):
            return eval_constraint(self, '<<', other)
        return NotImplemented

    def __le__(self, other):
        if type(self) is type(other):
            return eval_constraint(self, '<=', other)
        return NotImplemented

    def __gt__(self, other):
        if type(self) is type(other):
            return eval_constraint(self, '>>', other)
        return NotImplemented

    def __ge__(self, other):
        if type(self) is type(other):
            return eval_constraint(self, '>=', other)
        return NotImplemented

    @classmethod
    def from_string(cls, version):
        if not version and not isinstance(version , str):
            raise ValueError('Invalid version string: "{}"'.format(version))
        version = version.strip()
        if not version:
            raise ValueError('Invalid version string: "{}"'.format(version))
        if not  _is_valid_version(version):
            raise ValueError('Invalid version string: "{}"'.format(version))

        if ":" in version:
            epoch, _, version = version.partition(':')
            epoch = int(epoch)
        else:
            epoch = 0

        if "-" in version:
            upstream, _, revision = version.rpartition('-')
        else:
            upstream = version
            revision = '0'
        return cls(epoch=epoch, upstream=upstream, revision=revision)

    def compare(self, other_version):
        return compare_versions(self, other_version)

    def to_dict(self):
        return asdict(self)

    def tuple(self):
        return self.epoch, self.upstream, self.revision


_is_valid_version = re.compile(
    r'^'
    # epoch must start with a digit
    r'(\d+:)?'
    # upstream must start with a digit
    r'\d'
    r'('
      # upstream  can contain only alphanumerics and the characters . + -
      # ~ (full stop, plus, hyphen, tilde)
      # we are adding the extra check that it must end with alphanum
      r'[A-Za-z0-9\.\+\-\~]*[A-Za-z0-9]'
    r'|'
      # If there is no debian_revision then hyphens are not allowed.
      # we are adding the extra check that it must end with alphanum
      r'[A-Za-z0-9\.\+\~]*[A-Za-z0-9]-[A-Za-z0-9\+\.\~]*[A-Za-z0-9]'
    r')?'
    r'$').match


def eval_constraint(version1, operator, version2):
    """
    Evaluate a versions constraint where two Debian package versions are
    compared with an operator such as < or >. Return True if the constraint is
    satisfied and False otherwise.
    """

    version1 = coerce_version(version1)
    version2 = coerce_version(version2)

    result = compare_versions(version1, version2)
    # See https://www.debian.org/doc/debian-policy/ch-relationships.html#syntax-of-relationship-fields
    operators = {
        '<=': operator_module.le,
        # legacy for compat
        '<': operator_module.le,

        '>=': operator_module.ge,
        # legacy for compat
        '>': operator_module.ge,

        '<<': operator_module.lt,
        '>>': operator_module.gt,

        '=': operator_module.eq,
    }

    try:
        operator = operators[operator]
    except KeyError:
        msg = f'Unsupported Debian version constraint comparison operator: {version1} {operator} {version2}'
        raise ValueError(msg)
    return operator(result, 0)


def compare_versions_key(x):
    """
    Return a key version function suitable for use in sorted().
    """
    return cmp_to_key(compare_versions)(x)


def compare_strings_key(x):
    """
    Return a key string function suitable for use in sorted().
    """
    return cmp_to_key(compare_strings)(x)


def compare_strings(version1, version2):
    """
    Compare two version strings (upstream or revision) using Debain semantics
    and return one of the following integer numbers:

        - -1 means version1 sorts before version2
        - 0 means version1 and version2 are equal
        - 1 means version1 sorts after version2
    """
    logger.debug("Comparing Debian version number substrings %r and %r ..", version1, version2)
    mapping = characters_order
    v1 = list(version1)
    v2 = list(version2)
    while v1 or v2:
        # Quoting from the 'deb-version' manual page: First the initial part of each
        # string consisting entirely of non-digit characters is determined. These two
        # parts (one of which may be empty) are compared lexically. If a difference is
        # found it is returned. The lexical comparison is a comparison of ASCII values
        # modified so that all the letters sort earlier than all the non-letters and so
        # that a tilde sorts before anything, even the end of a part. For example, the
        # following parts are in sorted order: '~~', '~~a', '~', the empty part, 'a'.
        p1 = get_non_digit_prefix(v1)
        p2 = get_non_digit_prefix(v2)
        if p1 != p2:
            logger.debug("Comparing non-digit prefixes %r and %r ..", p1, p2)
            for c1, c2 in zip_longest(p1, p2, fillvalue=""):
                logger.debug("Performing lexical comparison between characters %r and %r ..", c1, c2)
                o1 = mapping.get(c1)
                o2 = mapping.get(c2)
                if o1 < o2:
                    logger.debug("Determined that %r sorts before %r (based on lexical comparison).", version1, version2)
                    return -1
                elif o1 > o2:
                    logger.debug("Determined that %r sorts after %r (based on lexical comparison).", version1, version2)
                    return 1
        elif p1:
            logger.debug("Skipping matching non-digit prefix %r ..", p1)
        # Quoting from the 'deb-version' manual page: Then the initial part of the
        # remainder of each string which consists entirely of digit characters is
        # determined. The numerical values of these two parts are compared, and any
        # difference found is returned as the result of the comparison. For these purposes
        # an empty string (which can only occur at the end of one or both version strings
        # being compared) counts as zero.
        d1 = get_digit_prefix(v1)
        d2 = get_digit_prefix(v2)
        logger.debug("Comparing numeric prefixes %i and %i ..", d1, d2)
        if d1 < d2:
            logger.debug("Determined that %r sorts before %r (based on numeric comparison).", version1, version2)
            return -1
        elif d1 > d2:
            logger.debug("Determined that %r sorts after %r (based on numeric comparison).", version1, version2)
            return 1
        else:
            logger.debug("Determined that numeric prefixes match.")
    logger.debug("Determined that version numbers are equal.")
    return 0


def compare_versions(version1, version2):
    """
    Compare two Version objects or strings and return one of the following
    integer numbers:

      - -1 means version1 sorts before version2
      - 0 means version1 and version2 are equal
      - 1 means version1 sorts after version2
    """
    version1 = coerce_version(version1)
    version2 = coerce_version(version2)
    return compare_version_objects(version1, version2)


def coerce_version(value):
    """
    Return a Version object from value.

    :param value: The value to coerce (a string or :class:`Version` object).
    :returns: A :class:`Version` object.
    """
    if not isinstance(value, Version):
        value = Version.from_string(value)
    return value


def compare_version_objects(version1, version2):
    """
    Compare two Version objects and return one of the following
    integer numbers:

      - -1 means version1 sorts before version2
      - 0 means version1 and version2 are equal
      - 1 means version1 sorts after version2
    """
    if version1.epoch < version2.epoch:
        return -1
    if version1.epoch > version2.epoch:
        return 1
    result = compare_strings(version1.upstream, version2.upstream)
    if result != 0:
        return result
    if version1.revision or version2.revision:
        return compare_strings(version1.revision, version2.revision)
    return 0


def get_digit_prefix(characters):
    """
    Return the digit prefix from a list of characters.
    """
    value = 0
    while characters and characters[0].isdigit():
        value = value * 10 + int(characters.pop(0))
    return value


def get_non_digit_prefix(characters):
    """
    Return the non-digit prefix from a list of characters.
    """
    prefix = []
    while characters and not characters[0].isdigit():
        prefix.append(characters.pop(0))
    return prefix


# a mapping of characters to integers representing the Debian sort order.
characters_order = {
    # The tilde sorts before everything.
    '~': 0,
    # The empty string sort before everything except a tilde.
    '': 1,
    # Letters sort before everything but a tilde or empty string, in their regular lexical sort order.
    'A': 2,
    'B': 3,
    'C': 4,
    'D': 5,
    'E': 6,
    'F': 7,
    'G': 8,
    'H': 9,
    'I': 10,
    'J': 11,
    'K': 12,
    'L': 13,
    'M': 14,
    'N': 15,
    'O': 16,
    'P': 17,
    'Q': 18,
    'R': 19,
    'S': 20,
    'T': 21,
    'U': 22,
    'V': 23,
    'W': 24,
    'X': 25,
    'Y': 26,
    'Z': 27,
    'a': 28,
    'b': 29,
    'c': 30,
    'd': 31,
    'e': 32,
    'f': 33,
    'g': 34,
    'h': 35,
    'i': 36,
    'j': 37,
    'k': 38,
    'l': 39,
    'm': 40,
    'n': 41,
    'o': 42,
    'p': 43,
    'q': 44,
    'r': 45,
    's': 46,
    't': 47,
    'u': 48,
    'v': 49,
    'w': 50,
    'x': 51,
    'y': 52,
    'z': 53,
    # Punctuation characters follow in their regular lexical sort order.
    '+': 54,
    '-': 55,
    '.': 56,
}



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/debian-inspector for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

from os import path

from test_utils import JsonTester  # NOQA

from debian_inspector import debcon


class TestGetParagraphData(JsonTester):
    test_data_dir = path.join(path.dirname(__file__), 'data')

    def test_get_paragraph_data_from_file_does_not_crash_on_None(self):
        results = list(debcon.get_paragraph_data_from_file(None))
        assert results == []

    def test_get_paragraph_data_from_file_from_single_status(self):
        test_file = self.get_test_loc('debcon/status/one_status')
        expected_loc = 'debcon/status/one_status-expected.json'
        results = debcon.get_paragraph_data_from_file(test_file)
        self.check_json(results, expected_loc, regen=False)

    def test_get_paragraph_data_from_file_from_status_with_junk_uses_unknown(self):
        test_file = self.get_test_loc('debcon/status/one_status_junk', must_exist=False)
        expected_loc = 'debcon/status/one_status_junk-expected.json'
        results = debcon.get_paragraph_data_from_file(test_file)
        self.check_json(results, expected_loc, regen=False)

    def test_get_paragraph_data_from_file__signed_from_dsc_can_remove_signature(self):
        test_file = self.get_test_loc('debcon/dsc/zlib_1.2.11.dfsg-1.dsc')
        expected_loc = 'debcon/dsc/zlib_1.2.11.dfsg-1.dsc-expected.json'
        results = debcon.get_paragraph_data_from_file(test_file, remove_pgp_signature=True)
        self.check_json(results, expected_loc, regen=False)

    def test_get_paragraph_data_from_file__signed_from_dsc_does_not_crash_if_signature_not_removed(self):
        test_file = self.get_test_loc('debcon/dsc/zlib_1.2.11.dfsg-1.dsc')
        expected_loc = 'debcon/dsc/zlib_1.2.11.dfsg-1.dsc-expected-no-desig.json'
        results = debcon.get_paragraph_data_from_file(test_file, remove_pgp_signature=False)
        self.check_json(results, expected_loc, regen=False)

    def test_get_paragraph_data_from_file_from_status_can_handle_perl_status(self):
        test_file = self.get_test_loc('debcon/status/perl_status')
        expected_loc = 'debcon/status/perl_status-expected.json'
        results = debcon.get_paragraph_data_from_file(test_file)
        self.check_json(results, expected_loc, regen=False)

    def test_get_paragraph_data__invalid_format_returns_none(self):
        text = '''Some text that is not RFC 822
        Compliant'''
        results = debcon.get_paragraph_data(text)
        assert results == {'unknown': text}

    def test_get_paragraph_data_does_not_preserve_keys_case_by_default(self):
        key = 'Some-text'
        value = '''RFC 822
        Compliant'''
        test = '{}: {}'.format(key, value)
        results = debcon.get_paragraph_data(test)
        assert results == {key.lower(): value}

    def test_get_paragraph_data__merge_duplicate_keys(self):
        o1 = 'some: val'
        kv1 = 'key: value1'
        o2 = 'other: val'
        kv2 = 'Key: value2'
        test = '{}\n{}\n{}\n{}'.format(o1, kv1, o2, kv2)
        results = debcon.get_paragraph_data(test)
        expected = [
            ('some', 'val'),
            ('key', 'value1\nvalue2'),
            ('other', 'val')]
        assert list(results.items()) == expected

    def test_test_get_paragraph_data__simple(self):
        items = 'A: b\nc: d'
        results = debcon.get_paragraph_data(items)
        expected = {'a': 'b', 'c': 'd'}
        assert results == expected

    def test_test_get_paragraph_data_lowers_only_keys(self):
        items = 'A: B\nDF: D'
        results = debcon.get_paragraph_data(items)
        expected = {'a': 'B', 'df': 'D'}
        assert results == expected

    def test_test_get_paragraph_data_merge_dupes(self):
        items = 'A: B\nDF: D\ndf: x'
        results = debcon.get_paragraph_data(items)
        expected = {'a': 'B', 'df': 'D\nx'}
        assert results == expected

    def test_get_paragraphs_data__splits_paragraphs_correctly(self):
        test = 'para1: test1\n\npara2: test2'
        results = list(debcon.get_paragraphs_data(test))
        expected = [{'para1': 'test1'}, {'para2': 'test2'}]
        assert results == expected

    def test_split_in_paragraphs__splits_paragraphs_correctly(self):
        test = 'para1: test1\n\npara2: test2'
        results = list(debcon.split_in_paragraphs(test))
        expected = ['para1: test1', 'para2: test2']
        assert results == expected

    def test_split_in_paragraphs__splits_paragraphs_with_multiple_lines_correctly(self):
        test_text_split = """Upstream-Name: GnuPG

Files: *
Copyright: Free Software Foundation, Inc
License: GPL-3+

License: TinySCHEME
 Redistribution


License: permissive
 This file is free software.

License: RFC-Reference
 doc/OpenPGP


License: GPL-3+
 GnuPG
"""

        results = list(debcon.split_in_paragraphs(test_text_split))
        expected = [
            'Upstream-Name: GnuPG',
            'Files: *\nCopyright: Free Software Foundation, Inc\nLicense: GPL-3+',
            'License: TinySCHEME\n Redistribution',
            'License: permissive\n This file is free software.',
            'License: RFC-Reference\n doc/OpenPGP',
            'License: GPL-3+\n GnuPG\n',
        ]
        assert results == expected

    def test_split_in_paragraphs__handles_more_than_two_empty_lines(self):
        test = 'para1: test1\n\n\n\n\npara2: test2'
        results = list(debcon.split_in_paragraphs(test))
        expected = ['para1: test1', 'para2: test2']
        assert results == expected

    def test_split_in_paragraphs__handles_empty_lines_with_spaces(self):
        test = 'para1: test1\n\n \t     \n          \npara2: test2'
        results = list(debcon.split_in_paragraphs(test))
        expected = ['para1: test1', 'para2: test2']
        assert results == expected

    def test_get_paragraphs_data_from_text__from_status_file(self):
        test_file = self.get_test_loc('debcon/status/simple_status')
        with open(test_file) as tf:
            test_file = tf.read()
        expected_loc = 'debcon/status/simple_status-expected.json'
        results = list(debcon.get_paragraphs_data(test_file))
        self.check_json(results, expected_loc, regen=False)

    def test_get_paragraphs_data_from_file__from_copyrights_dep5_1(self):
        test_file = self.get_test_loc('debcon/copyright/dep5-b43-fwcutter.copyright')
        expected_loc = 'debcon/copyright/dep5-b43-fwcutter.copyright-expected.json'
        results = list(debcon.get_paragraphs_data_from_file(test_file))
        self.check_json(results, expected_loc, regen=False)

    def test_get_paragraphs_data_from_file__from_copyrights_dep5_3(self):
        test_file = self.get_test_loc('debcon/copyright/dep5-rpm.copyright')
        expected_loc = 'debcon/copyright/dep5-rpm.copyright-expected.json'
        results = list(debcon.get_paragraphs_data_from_file(test_file))
        self.check_json(results, expected_loc, regen=False)

    def test_get_paragraphs_data_from_file__from_copyrights_dep5_dropbear(self):
        test_file = self.get_test_loc('debcon/copyright/dropbear.copyright')
        expected_loc = 'debcon/copyright/dropbear.copyright-expected.json'
        results = list(debcon.get_paragraphs_data_from_file(test_file))
        self.check_json(results, expected_loc, regen=False)

    def test_get_paragraphs_data_from_file__from_packages(self):
        test_file = self.get_test_loc('debcon/packages/simple_packages')
        expected_loc = 'debcon/packages/simple_packages-expected.json'
        results = list(debcon.get_paragraphs_data_from_file(test_file))
        self.check_json(results, expected_loc, regen=False)

    def test_get_paragraphs_data_from_file__from_sources(self):
        test_file = self.get_test_loc('debcon/sources/simple_sources')
        expected_loc = 'debcon/sources/simple_sources-expected.json'
        results = list(debcon.get_paragraphs_data_from_file(test_file))
        self.check_json(results, expected_loc, regen=False)

    def test_get_paragraphs_data_from_file__from_status(self):
        test_file = self.get_test_loc('debcon/status/simple_status')
        expected_loc = 'debcon/status/simple_status-expected.json'
        results = list(debcon.get_paragraphs_data_from_file(test_file))
        self.check_json(results, expected_loc, regen=False)

    def test_get_paragraph_data__return_unknow_if_we_have_payload(self):
        # we were skipping email payloads: a payload means this is not a
        # header only file and therefore something we cannot process normally
        test = 'Foo: home\n\nBar: baz'
        results = debcon.get_paragraph_data(test)
        expected = {'foo': 'home', 'unknown': 'Bar: baz'}
        assert results == expected


class TestDebian822(JsonTester):
    test_data_dir = path.join(path.dirname(__file__), 'data')

    def test_Debian822_from_file__from_one_status(self):
        test_file = self.get_test_loc('debcon/deb822/one_status')
        expected_loc = 'debcon/deb822/one_status-expected-deb822.json'
        results = debcon.Debian822.from_file(test_file).to_dict()
        self.check_json(results, expected_loc, regen=False)

    def test_Debian822_from_file__from_one_status_keeps_unknown_junk(self):
        test_file = self.get_test_loc('debcon/deb822/one_status_junk')
        expected_loc = 'debcon/deb822/one_status_junk-expected-deb822.json'
        results = debcon.Debian822.from_file(test_file).to_dict()
        self.check_json(results, expected_loc, regen=False)

    def test_Debian822_from_string__from_one_status(self):
        test_file = self.get_test_loc('debcon/deb822/one_status')
        with open(test_file) as tf:
            test_file = tf.read()
        expected_loc = 'debcon/deb822/one_status-expected-deb822.json'
        results = debcon.Debian822.from_string(test_file).to_dict()
        self.check_json(results, expected_loc, regen=False)

    def test_Debian822_from_file__signed_from_dsc(self):
        test_file = self.get_test_loc('debcon/deb822/zlib_1.2.11.dfsg-1.dsc')
        expected_loc = 'debcon/deb822/zlib_1.2.11.dfsg-1.dsc-expected-deb822.json'
        results = debcon.Debian822.from_file(test_file).to_dict()
        self.check_json(results, expected_loc, regen=False)

    def test_Debian822_from_items_list(self):
        items = [
            ('Package', 'debian_inspector'),
            ('Depends', 'python, python-pip, python-pip-accel'),
            ('Installed-Size', '65'),
        ]
        d822 = debcon.Debian822(items)
        expected = {
            'depends': 'python, python-pip, python-pip-accel',
            'installed-size': '65',
            'package': 'debian_inspector',
        }
        assert d822.to_dict() == expected
        expected2 = {
            'depends': 'python, python-pip, python-pip-accel',
            'installed-size': '65',
            'package': 'debian_inspector',
        }
        assert dict(d822) == expected2


class TestDebianFields(JsonTester):
    test_data_dir = path.join(path.dirname(__file__), 'data')

    def test_FormattedTextField(self):
        test = 'simple'
        results = debcon.FormattedTextField.from_value(test)
        assert results.text == 'simple'
        assert results.dumps() == 'simple'

        test = ' simple'
        results = debcon.FormattedTextField.from_value(test)
        assert results.text == 'simple'
        assert results.dumps() == 'simple'

        test = '  simple'
        results = debcon.FormattedTextField.from_value(test)
        assert results.text == 'simple'
        assert results.dumps() == 'simple'

    def test_FormattedTextField_multilines(self):
        test = ''' complex
 some
 .
  nostrip
 .
'''
        results = debcon.FormattedTextField.from_value(test)
        expected = 'complex\nsome\n\n nostrip\n'
        assert results.text == expected
        expected = 'complex\n some\n .\n  nostrip'
        assert results.dumps() == expected

    def test_DescriptionField(self):
        test = 'simple'
        results = debcon.DescriptionField.from_value(test)
        assert results.synopsis == 'simple'
        assert not results.text
        assert results.dumps() == 'simple'

        test = ' simple'
        assert results.synopsis == 'simple'
        assert not results.text
        assert results.dumps() == 'simple'

        test = '  simple'
        assert results.synopsis == 'simple'
        assert not results.text
        assert results.dumps() == 'simple'

    def test_DescriptionField_multilines(self):
        test = ''' complex
 some
 .
  nostrip
 .
'''
        results = debcon.DescriptionField.from_value(test)
        assert results.synopsis == 'complex'
        assert results.text == 'some\n\n nostrip\n'
        assert results.dumps() == 'complex\n some\n .\n  nostrip'

    def test_MaintainerField(self):
        test = ' Joe Z. Doe   <me@jzd.me> '
        results = debcon.MaintainerField.from_value(test)
        assert results.name == 'Joe Z. Doe'
        assert results.email_address == 'me@jzd.me'
        assert results.dumps() == 'Joe Z. Doe <me@jzd.me>'

    def test_MaintainerField_incorrect_email(self):
        test = ' Joe Z. Doe   me@j zd.me> '
        results = debcon.MaintainerField.from_value(test)
        assert results.name == 'Joe Z. Doe   me@j zd.me>'
        assert not results.email_address
        assert results.dumps() == 'Joe Z. Doe   me@j zd.me>'

    def test_SingleLineField(self):
        test = ' some value '
        results = debcon.SingleLineField.from_value(test)
        assert results.value == 'some value'
        assert results.dumps() == 'some value'
        assert results.dumps() == str(results)

    def test_LineSeparatedField(self):
        test = ' some value   \n   some value   \n     some more   '
        results = debcon.LineSeparatedField.from_value(test)
        assert results.values == ['some value', 'some value', 'some more']
        assert results.dumps() == 'some value\n some value\n some more'

    def test_AnyWhiteSpaceSeparatedField(self):
        test = ' some value   \n   some value   \n     some more   '
        results = debcon.AnyWhiteSpaceSeparatedField.from_value(test)
        assert results.values == ['some', 'value', 'some', 'value', 'some', 'more']
        assert results.dumps() == 'some\n value\n some\n value\n some\n more'

    def test_LineAndSpaceSeparatedField(self):
        test = ' some value \n some value   \n     some more'
        results = debcon.LineAndSpaceSeparatedField.from_value(test)
        assert results.values == [('some', 'value'), ('some', 'value'), ('some', 'more')]
        assert results.dumps() == 'some value\n some value\n some more'
